from docs.examples.request_data.msgpack_request import app as msgpack_app
from docs.examples.request_data.request_data_1 import app
from docs.examples.request_data.request_data_2 import app as app_2
from docs.examples.request_data.request_data_3 import app as app_3
from docs.examples.request_data.request_data_4 import app as app_4
from docs.examples.request_data.request_data_5 import app as app_5
from docs.examples.request_data.request_data_6 import app as app_6
from docs.examples.request_data.request_data_7 import app as app_7
from docs.examples.request_data.request_data_8 import app as app_8
from docs.examples.request_data.request_data_9 import app as app_9
from docs.examples.request_data.request_data_10 import app as app_10

from litestar.serialization import encode_msgpack
from litestar.testing import TestClient


def test_request_data_1() -> None:
    with TestClient(app=app) as client:
        response = client.post("/", json={"hello": "world"})
        assert response.status_code == 201
        assert response.json() == {"hello": "world"}


def test_request_data_2() -> None:
    with TestClient(app=app_2) as client:
        response = client.post("/", json={"id": 1, "name": "John"})
        assert response.status_code == 201
        assert response.json() == {"id": 1, "name": "John"}


def test_request_data_3() -> None:
    with TestClient(app=app_3) as client:
        response = client.post("/", json={"id": 1, "name": "John"})
        assert response.status_code == 201
        assert response.json() == {"id": 1, "name": "John"}


def test_request_data_4() -> None:
    with TestClient(app=app_4) as client:
        response = client.post("/", data={"id": 1, "name": "John"})
        assert response.status_code == 201
        assert response.json() == {"id": 1, "name": "John"}


def test_request_data_5() -> None:
    with TestClient(app=app_5) as client:
        response = client.post(
            "/",
            content=b'--d26a9a4ed2f441fba9ab42d04b42099e\r\nContent-Disposition: form-data; name="id"\r\n\r\n1\r\n--d26a9a4ed2f441fba9ab42d04b42099e\r\nContent-Disposition: form-data; name="name"\r\n\r\nJohn\r\n--d26a9a4ed2f441fba9ab42d04b42099e--\r\n',
            headers={
                "Content-Length": "211",
                "Content-Type": "multipart/form-data; boundary=d26a9a4ed2f441fba9ab42d04b42099e",
            },
        )
        assert response.json() == {"id": 1, "name": "John"}
        assert response.status_code == 201


def test_request_data_6() -> None:
    with TestClient(app=app_6) as client:
        response = client.post("/", files={"upload": ("hello", b"world")})
        assert response.status_code == 201
        assert response.text == "hello, world"


def test_request_data_7() -> None:
    with TestClient(app=app_7) as client:
        response = client.post("/", files={"upload": ("hello", b"world")})
        assert response.status_code == 201
        assert response.text == "hello, world"


def test_request_data_8() -> None:
    with TestClient(app=app_8) as client:
        response = client.post(
            "/", files={"cv": ("cv.odf", b"very impressive"), "diploma": ("diploma.pdf", b"the best")}
        )
        assert response.status_code == 201
        assert response.json() == {"cv": "very impressive", "diploma": "the best"}


def test_request_data_9() -> None:
    with TestClient(app=app_9) as client:
        response = client.post("/", files={"hello": b"there", "i'm": "steve"})
        assert response.status_code == 201
        assert response.json() == {"hello": "there", "i'm": "steve"}


def test_request_data_10() -> None:
    with TestClient(app=app_10) as client:
        response = client.post("/", files={"foo": ("foo.txt", b"hello"), "bar": ("bar.txt", b"world")})
        assert response.status_code == 201
        assert response.json() == {"foo.txt": "hello", "bar.txt": "world"}


def test_msgpack_app() -> None:
    test_data = {"name": "Moishe Zuchmir", "age": 30, "programmer": True}

    with TestClient(app=msgpack_app) as client:
        response = client.post("/", content=encode_msgpack(test_data))
        assert response.json() == test_data



from __future__ import annotations

from dataclasses import MISSING, fields, replace
from typing import TYPE_CHECKING, Generic, TypeVar

from litestar.dto.base_dto import AbstractDTO
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.dto.field import DTO_FIELD_META_KEY, DTOField
from litestar.params import DependencyKwarg, KwargDefinition
from litestar.types.empty import Empty

if TYPE_CHECKING:
    from typing import Collection, Generator

    from litestar.types.protocols import DataclassProtocol
    from litestar.typing import FieldDefinition


__all__ = ("DataclassDTO", "T")

T = TypeVar("T", bound="DataclassProtocol | Collection[DataclassProtocol]")
AnyDataclass = TypeVar("AnyDataclass", bound="DataclassProtocol")


class DataclassDTO(AbstractDTO[T], Generic[T]):
    """Support for domain modelling with dataclasses."""

    @classmethod
    def generate_field_definitions(
        cls, model_type: type[DataclassProtocol]
    ) -> Generator[DTOFieldDefinition, None, None]:
        dc_fields = {f.name: f for f in fields(model_type)}
        for key, field_definition in cls.get_model_type_hints(model_type).items():
            if not (dc_field := dc_fields.get(key)):
                continue

            default = dc_field.default if dc_field.default is not MISSING else Empty
            default_factory = dc_field.default_factory if dc_field.default_factory is not MISSING else None
            field_defintion = replace(
                DTOFieldDefinition.from_field_definition(
                    field_definition=field_definition,
                    default_factory=default_factory,
                    dto_field=dc_field.metadata.get(DTO_FIELD_META_KEY, DTOField()),
                    model_name=model_type.__name__,
                ),
                name=key,
                default=default,
            )

            yield (
                replace(field_defintion, default=Empty, kwarg_definition=default)
                if isinstance(default, (KwargDefinition, DependencyKwarg))
                else field_defintion
            )

    @classmethod
    def detect_nested_field(cls, field_definition: FieldDefinition) -> bool:
        return hasattr(field_definition.annotation, "__dataclass_fields__")



from __future__ import annotations

from datetime import date, datetime
from typing import Protocol, Type, Union, cast
from uuid import uuid4

import pytest
from _pytest.fixtures import FixtureRequest
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from litestar.contrib.sqlalchemy import base
from litestar.repository.exceptions import ConflictError, RepositoryError
from litestar.repository.filters import LimitOffset
from litestar.repository.testing.generic_mock_repository import (
    GenericAsyncMockRepository,
    GenericSyncMockRepository,
)
from tests.helpers import maybe_async
from tests.unit.test_repository.models_uuid import UUIDAuthor, UUIDBook

AuthorRepository = GenericAsyncMockRepository[UUIDAuthor]
AuthorRepositoryType = Type[AuthorRepository]
ModelType = Type[Union[base.UUIDBase, base.BigIntBase]]
AuditModelType = Type[Union[base.UUIDAuditBase, base.BigIntAuditBase]]


class CreateAuditModelFixture(Protocol):
    def __call__(self, extra_columns: dict[str, type[Mapped] | Mapped] | None = None) -> AuditModelType:
        ...


@pytest.fixture(name="authors")
def fx_authors() -> list[UUIDAuthor]:
    """Collection of Author models."""
    return [
        UUIDAuthor(id=uuid4(), name=name, dob=dob, created_at=datetime.min, updated_at=datetime.min)
        for name, dob in [("Agatha Christie", date(1890, 9, 15)), ("Leo Tolstoy", date(1828, 9, 9))]
    ]


@pytest.fixture(params=[GenericAsyncMockRepository, GenericSyncMockRepository], ids=["async", "sync"])
def repository_type(request: FixtureRequest) -> type[GenericAsyncMockRepository]:
    return cast("type[GenericAsyncMockRepository]", request.param)


@pytest.fixture(name="author_repository_type", params=[GenericAsyncMockRepository, GenericSyncMockRepository])
def fx_author_repository_type(
    authors: list[UUIDAuthor], monkeypatch: pytest.MonkeyPatch, repository_type: type[GenericAsyncMockRepository]
) -> AuthorRepositoryType:
    """Mock Author repository, pre-seeded with collection data."""
    repo = repository_type[UUIDAuthor]  # type: ignore[index]
    repo.seed_collection(authors)
    return cast("type[GenericAsyncMockRepository]", repo)


@pytest.fixture(name="author_repository")
def fx_author_repository(
    author_repository_type: type[GenericAsyncMockRepository[UUIDAuthor]],
) -> GenericAsyncMockRepository[UUIDAuthor]:
    """Mock Author repository instance."""
    return author_repository_type()


@pytest.fixture(params=[base.UUIDBase, base.BigIntBase])
def model_type(request: FixtureRequest) -> ModelType:
    return cast(ModelType, type(f"{request.node.nodeid}Model", (request.param,), {}))


@pytest.fixture(params=[base.UUIDAuditBase, base.BigIntAuditBase])
def create_audit_model_type(request: FixtureRequest) -> CreateAuditModelFixture:
    def create(extra_columns: dict[str, type[Mapped] | Mapped] | None = None) -> AuditModelType:
        return cast(AuditModelType, type(f"{request.node.nodeid}AuditModel", (request.param,), extra_columns or {}))

    return create


@pytest.fixture()
def audit_model_type(create_audit_model_type: CreateAuditModelFixture) -> AuditModelType:
    return create_audit_model_type()


async def test_repo_raises_conflict_if_add_with_id(
    authors: list[UUIDAuthor], author_repository: AuthorRepository
) -> None:
    """Test mock repo raises conflict if add identified entity."""
    with pytest.raises(ConflictError):
        await maybe_async(author_repository.add(authors[0]))


async def test_repo_raises_conflict_if_add_many_with_id(
    authors: list[UUIDAuthor], author_repository: AuthorRepository
) -> None:
    """Test mock repo raises conflict if add identified entity."""
    with pytest.raises(ConflictError):
        await maybe_async(author_repository.add_many(authors))


def test_generic_mock_repository_parametrization(repository_type: type[GenericAsyncMockRepository]) -> None:
    """Test that the mock repository handles multiple types."""
    author_repo = repository_type[UUIDAuthor]  # type: ignore[index]
    book_repo = repository_type[UUIDBook]  # type: ignore[index]
    assert author_repo.model_type is UUIDAuthor
    assert book_repo.model_type is UUIDBook


def test_generic_mock_repository_seed_collection(author_repository_type: AuthorRepositoryType) -> None:
    """Test seeding instances."""
    author_repository_type.seed_collection([UUIDAuthor(id="abc")])
    assert "abc" in author_repository_type.collection


def test_generic_mock_repository_clear_collection(author_repository_type: AuthorRepositoryType) -> None:
    """Test clearing collection for type."""
    author_repository_type.clear_collection()
    assert not author_repository_type.collection


def test_generic_mock_repository_filter_collection_by_kwargs(author_repository: AuthorRepository) -> None:
    """Test filtering the repository collection by kwargs."""
    collection = author_repository.filter_collection_by_kwargs(author_repository.collection, name="Leo Tolstoy")
    assert len(collection) == 1
    assert next(iter(collection.values())).name == "Leo Tolstoy"


def test_generic_mock_repository_filter_collection_by_kwargs_and_semantics(author_repository: AuthorRepository) -> None:
    """Test that filtering by kwargs has `AND` semantics when multiple kwargs,
    not `OR`."""
    collection = author_repository.filter_collection_by_kwargs(
        author_repository.collection, name="Agatha Christie", dob="1828-09-09"
    )
    assert len(collection) == 0


def test_generic_mock_repository_raises_repository_exception_if_named_attribute_doesnt_exist(
    author_repository: AuthorRepository,
) -> None:
    """Test that a repo exception is raised if a named attribute doesn't
    exist."""
    with pytest.raises(RepositoryError):
        _ = author_repository.filter_collection_by_kwargs(author_repository.collection, cricket="ball")


async def test_sets_created_updated_on_add(
    repository_type: type[GenericAsyncMockRepository], audit_model_type: AuditModelType
) -> None:
    """Test that the repository updates the 'created_at' and 'updated_at' timestamps
    if necessary."""

    instance = audit_model_type()
    assert "created_at" not in vars(instance)
    assert "updated_at" not in vars(instance)

    instance = await maybe_async(repository_type[audit_model_type]().add(instance))  # type: ignore[index]
    assert "created_at" in vars(instance)
    assert "updated_at" in vars(instance)


async def test_sets_updated_on_update(author_repository: AuthorRepository) -> None:
    """Test that the repository updates the 'updated' timestamp if
    necessary."""

    instance = next(iter(author_repository.collection.values()))
    original_updated = instance.updated_at
    instance = await maybe_async(author_repository.update(instance))
    assert instance.updated_at > original_updated


async def test_does_not_set_created_updated(
    repository_type: type[GenericAsyncMockRepository], model_type: ModelType
) -> None:
    """Test that the repository does not update the 'updated' timestamps when
    appropriate."""

    instance = model_type()
    repo = repository_type[model_type]()  # type: ignore[index]
    assert "created_at" not in vars(instance)
    assert "updated_at" not in vars(instance)

    instance = await maybe_async(repo.add(instance))
    assert "created_at" not in vars(instance)
    assert "updated_at" not in vars(instance)

    instance = await maybe_async(repo.update(instance))
    assert "created_at" not in vars(instance)
    assert "updated_at" not in vars(instance)


async def test_add(repository_type: type[GenericAsyncMockRepository], model_type: ModelType) -> None:
    """Test that the repository add method works correctly."""

    instance = model_type()

    inserted_uuid_instance = await maybe_async(repository_type[model_type]().add(instance))  # type: ignore[index]
    assert inserted_uuid_instance == instance


async def test_add_many(repository_type: type[GenericAsyncMockRepository], model_type: ModelType) -> None:
    """Test that the repository add_many method works correctly."""

    instances = [model_type(), model_type()]

    inserted_uuid_instances = await maybe_async(repository_type[model_type]().add_many(instances))  # type: ignore[index]

    assert len(instances) == len(inserted_uuid_instances)


async def test_update(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository update method works correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    mock_repo = repository_type[Model]()  # type: ignore[index]

    instance = await maybe_async(mock_repo.add(Model(random_column="A")))
    instance.random_column = "B"
    updated_instance = await maybe_async(mock_repo.update(instance))

    assert updated_instance == instance


async def test_update_many(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository add_many method works correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    mock_repo = repository_type[Model]()  # type: ignore[index]
    instances = [Model(random_column="A"), Model(random_column="B")]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    for instance in inserted_instances:
        instance.random_column = "C"
    updated_instances = await maybe_async(mock_repo.update_many(instances))
    for instance in updated_instances:
        assert instance.random_column == "C"
    assert len(instances) == len(updated_instances)


async def test_upsert(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository upsert method works correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    mock_repo = repository_type[Model]()  # type: ignore[index]

    instance = await maybe_async(mock_repo.upsert(Model(random_column="A")))
    instance.random_column = "B"
    updated_instance = await maybe_async(mock_repo.upsert(instance))

    assert updated_instance == instance


async def test_upsert_many(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository upsert method works correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    mock_repo = repository_type[Model]()  # type: ignore[index]

    instance = await maybe_async(mock_repo.upsert(Model(random_column="A")))
    instance.random_column = "B"
    new_instance = Model(random_column="C")
    updated_instances = await maybe_async(mock_repo.upsert_many([instance, new_instance]))

    assert new_instance in updated_instances
    assert instance in updated_instances


async def test_list(repository_type: type[GenericAsyncMockRepository], audit_model_type: AuditModelType) -> None:
    """Test that the repository list returns records."""

    mock_repo = repository_type[audit_model_type]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many([audit_model_type(), audit_model_type()]))
    listed_instances = await maybe_async(mock_repo.list())
    assert inserted_instances == listed_instances


async def test_delete(repository_type: type[GenericAsyncMockRepository], audit_model_type: AuditModelType) -> None:
    """Test that the repository delete functionality."""

    mock_repo = repository_type[audit_model_type]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many([audit_model_type(), audit_model_type()]))
    delete_instance = await maybe_async(mock_repo.delete(inserted_instances[0].id))
    assert delete_instance.id == inserted_instances[0].id
    count = await maybe_async(mock_repo.count())
    assert count == 1


async def test_delete_many(repository_type: type[GenericAsyncMockRepository], audit_model_type: AuditModelType) -> None:
    """Test that the repository delete many functionality."""

    mock_repo = repository_type[audit_model_type]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many([audit_model_type(), audit_model_type()]))
    delete_instances = await maybe_async(mock_repo.delete_many([obj.id for obj in inserted_instances]))
    assert len(delete_instances) == 2
    count = await maybe_async(mock_repo.count())
    assert count == 0


async def test_list_and_count(
    repository_type: type[GenericAsyncMockRepository], audit_model_type: AuditModelType
) -> None:
    """Test that the repository list_and_count returns records and the total record count."""

    instances = [audit_model_type(), audit_model_type()]
    mock_repo = repository_type[audit_model_type]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    listed_instances, count = await maybe_async(mock_repo.list_and_count())
    assert inserted_instances == listed_instances
    assert count == len(instances)


async def test_exists(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository exists returns booleans."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    instances = [Model(random_column="value 1"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    _ = await maybe_async(mock_repo.add_many(instances))
    exists = await maybe_async(mock_repo.exists(random_column="value 1"))
    assert exists


async def test_exists_with_filter(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository exists returns booleans. with filter argument"""
    limit_filter = LimitOffset(limit=1, offset=0)

    Model = create_audit_model_type({"random_column": Mapped[str]})

    instances = [Model(random_column="value 1"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    _ = await maybe_async(mock_repo.add_many(instances))
    exists = await maybe_async(mock_repo.exists(limit_filter, random_column="value 1"))
    assert exists


async def test_count(repository_type: type[GenericAsyncMockRepository], audit_model_type: AuditModelType) -> None:
    """Test that the repository count returns the total record count."""

    instances = [audit_model_type(), audit_model_type()]
    mock_repo = repository_type[audit_model_type]()  # type: ignore[index]
    _ = await maybe_async(mock_repo.add_many(instances))
    count = await maybe_async(mock_repo.count())
    assert count == len(instances)


async def test_get(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository get returns a model record correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    instances = [Model(random_column="value 1"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    item_id = inserted_instances[0].id
    fetched_instance = await maybe_async(mock_repo.get(item_id))
    assert inserted_instances[0] == fetched_instance


async def test_get_one(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository get_one returns a model record correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    instances = [Model(random_column="value 1"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    fetched_instance = await maybe_async(mock_repo.get_one(random_column="value 1"))
    assert inserted_instances[0] == fetched_instance
    with pytest.raises(RepositoryError):
        _ = await maybe_async(mock_repo.get_one(random_column="value 3"))


async def test_get_one_or_none(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository get_one_or_none returns a model record correctly."""

    Model = create_audit_model_type({"random_column": Mapped[str]})

    instances = [Model(random_column="value 1"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    fetched_instance = await maybe_async(mock_repo.get_one_or_none(random_column="value 1"))
    assert inserted_instances[0] == fetched_instance
    none_instance = await maybe_async(mock_repo.get_one_or_none(random_column="value 3"))
    assert none_instance is None


async def test_get_or_create(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository get_or_create returns a model record correctly."""

    Model = create_audit_model_type(
        {"random_column": Mapped[str], "cool_attribute": mapped_column(String, nullable=True)}
    )

    instances = [Model(random_column="value 1", cool_attribute="yep"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    fetched_instance, fetched_created = await maybe_async(mock_repo.get_or_create(random_column="value 2"))
    assert await maybe_async(mock_repo.count()) == 2
    assert inserted_instances[1] == fetched_instance
    assert fetched_created is False
    _, created = await maybe_async(mock_repo.get_or_create(random_column="value 3"))
    assert await maybe_async(mock_repo.count()) == 3
    assert created


async def test_get_or_create_match_fields(
    repository_type: type[GenericAsyncMockRepository], create_audit_model_type: CreateAuditModelFixture
) -> None:
    """Test that the repository get_or_create returns a model record correctly."""

    Model = create_audit_model_type(
        {"random_column": Mapped[str], "cool_attribute": mapped_column(String, nullable=True)}
    )

    instances = [Model(random_column="value 1", cool_attribute="yep"), Model(random_column="value 2")]
    mock_repo = repository_type[Model]()  # type: ignore[index]
    inserted_instances = await maybe_async(mock_repo.add_many(instances))
    fetched_instance, fetched_created = await maybe_async(
        mock_repo.get_or_create(match_fields=["random_column"], random_column="value 1", cool_attribute="other thing")
    )
    assert await maybe_async(mock_repo.count()) == 2
    assert inserted_instances[0] == fetched_instance
    assert fetched_created is False



from __future__ import annotations

from typing import Callable, Sequence, TypeVar

__all__ = ("find_index", "unique")


T = TypeVar("T")


def find_index(target_list: Sequence[T], predicate: Callable[[T], bool]) -> int:
    """Find element in list given a key and value.

    List elements can be dicts or classes
    """
    return next((i for i, element in enumerate(target_list) if predicate(element)), -1)


def unique(value: Sequence[T]) -> list[T]:
    """Return all unique values in a given sequence or iterator."""
    try:
        return list(set(value))
    except TypeError:
        output: list[T] = []
        for element in value:
            if element not in output:
                output.append(element)
        return output



from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import Any, AsyncGenerator, Dict, List, Optional, Type, Union, cast
from unittest.mock import AsyncMock, MagicMock

import pytest
from pytest_lazyfixture import lazy_fixture

from litestar import Controller, Litestar, Request, WebSocket
from litestar.datastructures import State
from litestar.di import Provide
from litestar.dto import DataclassDTO, dto_field
from litestar.exceptions import ImproperlyConfiguredException
from litestar.handlers.websocket_handlers import WebsocketListener, websocket_listener
from litestar.testing import create_test_client
from litestar.types.asgi_types import WebSocketMode


@pytest.fixture
def listener_class(mock: MagicMock) -> Type[WebsocketListener]:
    class Listener(WebsocketListener):
        def on_receive(self, data: str) -> str:
            mock(data)
            return data

    return Listener


@pytest.fixture
def sync_listener_callable(mock: MagicMock) -> websocket_listener:
    def listener(data: str) -> str:
        mock(data)
        return data

    return websocket_listener("/")(listener)


@pytest.fixture
def async_listener_callable(mock: MagicMock) -> websocket_listener:
    async def listener(data: str) -> str:
        mock(data)
        return data

    return websocket_listener("/")(listener)


@pytest.mark.parametrize(
    "listener",
    [
        lazy_fixture("sync_listener_callable"),
        lazy_fixture("async_listener_callable"),
        lazy_fixture("listener_class"),
    ],
)
def test_basic_listener(mock: MagicMock, listener: Union[websocket_listener, Type[WebsocketListener]]) -> None:
    client = create_test_client([listener])
    with client.websocket_connect("/") as ws:
        ws.send_text("foo")
        assert ws.receive_text() == "foo"
        ws.send_text("bar")
        assert ws.receive_text() == "bar"

    assert mock.call_count == 2
    mock.assert_any_call("foo")
    mock.assert_any_call("bar")


@pytest.mark.parametrize("receive_mode", ["text", "binary"])
def test_listener_receive_bytes(receive_mode: WebSocketMode, mock: MagicMock) -> None:
    @websocket_listener("/", receive_mode=receive_mode)
    def handler(data: bytes) -> None:
        mock(data)

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send("foo", mode=receive_mode)

    mock.assert_called_once_with(b"foo")


@pytest.mark.parametrize("receive_mode", ["text", "binary"])
def test_listener_receive_string(receive_mode: WebSocketMode, mock: MagicMock) -> None:
    @websocket_listener("/", receive_mode=receive_mode)
    def handler(data: str) -> None:
        mock(data)

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send("foo", mode=receive_mode)

    mock.assert_called_once_with("foo")


@pytest.mark.parametrize("receive_mode", ["text", "binary"])
def test_listener_receive_json(receive_mode: WebSocketMode, mock: MagicMock) -> None:
    @websocket_listener("/", receive_mode=receive_mode)
    def handler(data: List[str]) -> None:
        mock(data)

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_json(["foo", "bar"], mode=receive_mode)

    mock.assert_called_once_with(["foo", "bar"])


@dataclass
class User:
    name: str
    hidden: str = field(default="super secret", metadata=dto_field("private"))


@pytest.mark.parametrize("receive_mode", ["text", "binary"])
def test_listener_receive_with_dto(receive_mode: WebSocketMode) -> None:
    user_dto = DataclassDTO[User]
    value: Any = None

    @websocket_listener("/", receive_mode=receive_mode, dto=user_dto, return_dto=None)
    def handler(data: User) -> None:
        nonlocal value
        value = data

    client = create_test_client([handler], openapi_config=None)
    with client.websocket_connect("/") as ws:
        ws.send_json({"name": "litestar user", "hidden": "whoops"}, mode=receive_mode)

    assert isinstance(value, User)
    assert value.name == "litestar user"
    assert value.hidden == "super secret"


@pytest.mark.parametrize("send_mode", ["text", "binary"])
def test_listener_return_bytes(send_mode: WebSocketMode) -> None:
    @websocket_listener("/", send_mode=send_mode)
    def handler(data: str) -> bytes:
        return data.encode("utf-8")

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_text("foo")
        if send_mode == "text":
            assert ws.receive_text() == "foo"
        else:
            assert ws.receive_bytes() == b"foo"


@pytest.mark.parametrize("send_mode", ["text", "binary"])
def test_listener_send_json(send_mode: WebSocketMode) -> None:
    @websocket_listener("/", send_mode=send_mode)
    def handler(data: str) -> Dict[str, str]:
        return {"data": data}

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_text("foo")
        assert ws.receive_json(mode=send_mode) == {"data": "foo"}


@pytest.mark.parametrize("send_mode", ["text", "binary"])
def test_listener_send_with_dto(send_mode: WebSocketMode, mock: MagicMock) -> None:
    @dataclass
    class User:
        name: str
        hidden: str = field(default="super secret", metadata=dto_field("private"))

    user_dto = DataclassDTO[User]

    @websocket_listener("/", send_mode=send_mode, dto=user_dto, signature_namespace={"User": User})
    def handler(data: User) -> User:
        return data

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_json({"name": "litestar user"})
        assert ws.receive_json(mode=send_mode) == {"name": "litestar user"}


def test_listener_return_none() -> None:
    @websocket_listener("/")
    def handler(data: str) -> None:
        return data  # type: ignore[return-value]

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_text("foo")


def test_listener_return_optional_none() -> None:
    @websocket_listener("/")
    def handler(data: str) -> Optional[str]:
        return "world" if data == "hello" else None

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_text("hello")
        assert ws.receive_text() == "world"
        ws.send_text("goodbye")


def test_listener_pass_socket(mock: MagicMock) -> None:
    @websocket_listener("/")
    def handler(data: str, socket: WebSocket) -> Dict[str, str]:
        mock(socket=socket)
        return {"data": data}

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_text("foo")
        assert ws.receive_json() == {"data": "foo"}

    assert isinstance(mock.call_args.kwargs["socket"], WebSocket)


def test_listener_pass_additional_dependencies(mock: MagicMock) -> None:
    async def foo_dependency(state: State) -> int:
        if not hasattr(state, "foo"):
            state.foo = 0
        state.foo += 1
        return cast("int", state.foo)

    @websocket_listener("/", dependencies={"foo": Provide(foo_dependency)})
    def handler(data: str, foo: int) -> Dict[str, Union[str, int]]:
        return {"data": data, "foo": foo}

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        ws.send_text("something")
        ws.send_text("something")
        assert ws.receive_json() == {"data": "something", "foo": 1}


def test_listener_callback_no_data_arg_raises() -> None:
    with pytest.raises(ImproperlyConfiguredException):

        @websocket_listener("/")
        def handler() -> None:
            ...

        handler.on_registration(Litestar())


def test_listener_callback_request_and_body_arg_raises() -> None:
    with pytest.raises(ImproperlyConfiguredException):

        @websocket_listener("/")
        def handler_request(data: str, request: Request) -> None:
            ...

        handler_request.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @websocket_listener("/")
        def handler_body(data: str, body: bytes) -> None:
            ...

        handler_body.on_registration(Litestar())


def test_listener_accept_connection_callback() -> None:
    async def accept_connection(socket: WebSocket) -> None:
        await socket.accept(headers={"Cookie": "custom-cookie"})

    @websocket_listener("/", connection_accept_handler=accept_connection)
    def handler(data: bytes) -> None:
        return None

    client = create_test_client([handler])
    with client.websocket_connect("/") as ws:
        assert ws.extra_headers == [(b"cookie", b"custom-cookie")]


def test_connection_callbacks() -> None:
    on_accept_mock = MagicMock()
    on_disconnect_mock = MagicMock()

    def on_accept(socket: WebSocket) -> None:
        on_accept_mock()

    def on_disconnect(socket: WebSocket) -> None:
        on_disconnect_mock()

    @websocket_listener("/", on_accept=on_accept, on_disconnect=on_disconnect)
    def handler(data: bytes) -> None:
        pass

    client = create_test_client([handler])
    with client.websocket_connect("/"):
        pass

    on_accept_mock.assert_called_once()
    on_disconnect_mock.assert_called_once()


def test_connection_lifespan() -> None:
    on_accept = MagicMock()
    on_disconnect = MagicMock()

    @asynccontextmanager
    async def lifespan(socket: WebSocket) -> AsyncGenerator[None, None]:
        on_accept(socket)
        try:
            yield
        finally:
            on_disconnect(socket)

    @websocket_listener("/", connection_lifespan=lifespan)
    def handler(data: bytes) -> None:
        pass

    client = create_test_client([handler])
    with client.websocket_connect("/", timeout=1):
        pass

    on_accept.assert_called_once()
    on_disconnect.assert_called_once()


def test_listener_in_controller() -> None:
    # test for https://github.com/litestar-org/litestar/issues/1615

    class ClientController(Controller):
        path: str = "/"

        @websocket_listener("/ws")
        async def websocket_handler(self, data: str, socket: WebSocket) -> str:
            return data

    with create_test_client(ClientController) as client, client.websocket_connect("/ws") as ws:
        ws.send_text("foo")
        data = ws.receive_text(timeout=1)
        assert data == "foo"


def test_lifespan_dependencies() -> None:
    mock = MagicMock()

    @asynccontextmanager
    async def lifespan(name: str, state: State, query: dict) -> AsyncGenerator[None, None]:
        mock(name=name, state=state, query=query)
        yield

    @websocket_listener("/{name:str}", connection_lifespan=lifespan)
    async def handler(data: str) -> None:
        pass

    with create_test_client([handler]) as client, client.websocket_connect("/foo") as ws:
        ws.send_text("")

    assert mock.call_args_list[0].kwargs["name"] == "foo"
    assert isinstance(mock.call_args_list[0].kwargs["state"], State)
    assert isinstance(mock.call_args_list[0].kwargs["query"], dict)


def test_hook_dependencies() -> None:
    on_accept_mock = MagicMock()
    on_disconnect_mock = MagicMock()

    def some_dependency() -> str:
        return "hello"

    def on_accept(name: str, state: State, query: dict, some: str) -> None:
        on_accept_mock(name=name, state=state, query=query, some=some)

    def on_disconnect(name: str, state: State, query: dict, some: str) -> None:
        on_disconnect_mock(name=name, state=state, query=query, some=some)

    @websocket_listener("/{name: str}", on_accept=on_accept, on_disconnect=on_disconnect)
    def handler(data: bytes) -> None:
        pass

    with create_test_client([handler], dependencies={"some": some_dependency}) as client, client.websocket_connect(
        "/foo"
    ) as ws:
        ws.send_text("")

    on_accept_kwargs = on_accept_mock.call_args_list[0].kwargs
    assert on_accept_kwargs["name"] == "foo"
    assert on_accept_kwargs["some"] == "hello"
    assert isinstance(on_accept_kwargs["state"], State)
    assert isinstance(on_accept_kwargs["query"], dict)

    on_disconnect_kwargs = on_disconnect_mock.call_args_list[0].kwargs
    assert on_disconnect_kwargs["name"] == "foo"
    assert on_disconnect_kwargs["some"] == "hello"
    assert isinstance(on_disconnect_kwargs["state"], State)
    assert isinstance(on_disconnect_kwargs["query"], dict)


def test_websocket_listener_class_hook_dependencies() -> None:
    on_accept_mock = MagicMock()
    on_disconnect_mock = MagicMock()

    def some_dependency() -> str:
        return "hello"

    class Listener(WebsocketListener):
        path = "/{name: str}"

        def on_accept(self, name: str, state: State, query: dict, some: str) -> None:  # type: ignore[override]
            on_accept_mock(name=name, state=state, query=query, some=some)

        def on_disconnect(self, name: str, state: State, query: dict, some: str) -> None:  # type: ignore[override]
            on_disconnect_mock(name=name, state=state, query=query, some=some)

        def on_receive(self, data: bytes) -> None:
            pass

    with create_test_client([Listener], dependencies={"some": some_dependency}) as client, client.websocket_connect(
        "/foo"
    ) as ws:
        ws.send_text("")

    on_accept_kwargs = on_accept_mock.call_args_list[0].kwargs
    assert on_accept_kwargs["name"] == "foo"
    assert on_accept_kwargs["some"] == "hello"
    assert isinstance(on_accept_kwargs["state"], State)
    assert isinstance(on_accept_kwargs["query"], dict)

    on_disconnect_kwargs = on_disconnect_mock.call_args_list[0].kwargs
    assert on_disconnect_kwargs["name"] == "foo"
    assert on_disconnect_kwargs["some"] == "hello"
    assert isinstance(on_disconnect_kwargs["state"], State)
    assert isinstance(on_disconnect_kwargs["query"], dict)


@pytest.mark.parametrize("hook_name", ["on_accept", "on_disconnect", "connection_accept_handler"])
def test_listeners_lifespan_hooks_and_manager_raises(hook_name: str) -> None:
    @asynccontextmanager
    async def lifespan() -> AsyncGenerator[None, None]:
        yield

    hook_callback = AsyncMock()

    with pytest.raises(ImproperlyConfiguredException):

        @websocket_listener("/", **{hook_name: hook_callback}, connection_lifespan=lifespan)  # pyright: ignore
        def handler(data: bytes) -> None:
            pass



import logging
import sys
from typing import TYPE_CHECKING, Any, Dict
from unittest.mock import Mock, patch

import pytest

from litestar import Request, get
from litestar.logging.config import LoggingConfig, _get_default_handlers, default_handlers, default_picologging_handlers
from litestar.logging.picologging import QueueListenerHandler as PicologgingQueueListenerHandler
from litestar.logging.standard import QueueListenerHandler as StandardQueueListenerHandler
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


@pytest.mark.parametrize(
    "dict_config_class, handlers, expected_called",
    [
        ["logging.config.dictConfig", default_handlers, True],
        ["logging.config.dictConfig", default_picologging_handlers, False],
        ["picologging.config.dictConfig", default_handlers, False],
        ["picologging.config.dictConfig", default_picologging_handlers, True],
    ],
)
def test_correct_dict_config_called(
    dict_config_class: str, handlers: Dict[str, Dict[str, Any]], expected_called: bool
) -> None:
    with patch(dict_config_class) as dict_config_mock:
        log_config = LoggingConfig(handlers=handlers)
        log_config.configure()
        if expected_called:
            assert dict_config_mock.called
        else:
            assert not dict_config_mock.called


@pytest.mark.parametrize("picologging_exists", [True, False])
def test_correct_default_handlers_set(picologging_exists: bool) -> None:
    with patch("litestar.logging.config.find_spec") as find_spec_mock:
        find_spec_mock.return_value = picologging_exists
        log_config = LoggingConfig()

        if picologging_exists:
            assert log_config.handlers == default_picologging_handlers
        else:
            assert log_config.handlers == default_handlers


@pytest.mark.parametrize(
    "dict_config_class, handlers",
    [
        ["logging.config.dictConfig", default_handlers],
        ["picologging.config.dictConfig", default_picologging_handlers],
    ],
)
def test_dictconfig_startup(dict_config_class: str, handlers: Any) -> None:
    with patch(dict_config_class) as dict_config_mock:
        test_logger = LoggingConfig(
            handlers=handlers,
        )
        with create_test_client([], on_startup=[test_logger.configure]):
            assert dict_config_mock.called


def test_standard_queue_listener_logger(caplog: "LogCaptureFixture") -> None:
    with caplog.at_level("INFO", logger="test_logger"):
        logger = logging.getLogger("test_logger")
        logger.info("Testing now!")
        assert "Testing now!" in caplog.text
        var = "test_var"
        logger.info("%s", var)
        assert var in caplog.text


@patch("picologging.config.dictConfig")
def test_picologging_dictconfig_when_disabled(dict_config_mock: Mock) -> None:
    test_logger = LoggingConfig(loggers={"app": {"level": "INFO", "handlers": ["console"]}}, handlers=default_handlers)
    with create_test_client([], on_startup=[test_logger.configure], logging_config=None):
        assert not dict_config_mock.called


def test_get_default_logger() -> None:
    with create_test_client(logging_config=LoggingConfig(handlers=default_handlers)) as client:
        assert isinstance(client.app.logger.handlers[0], StandardQueueListenerHandler)
        new_logger = client.app.get_logger()
        assert isinstance(new_logger.handlers[0], StandardQueueListenerHandler)


def test_get_picologging_logger() -> None:
    with create_test_client(logging_config=LoggingConfig(handlers=default_picologging_handlers)) as client:
        assert isinstance(client.app.logger.handlers[0], PicologgingQueueListenerHandler)
        new_logger = client.app.get_logger()
        assert isinstance(new_logger.handlers[0], PicologgingQueueListenerHandler)


@pytest.mark.parametrize(
    "handlers, listener",
    [
        [default_handlers, StandardQueueListenerHandler],
        [default_picologging_handlers, PicologgingQueueListenerHandler],
    ],
)
def test_connection_logger(handlers: Any, listener: Any) -> None:
    @get("/")
    def handler(request: Request) -> Dict[str, bool]:
        return {"isinstance": isinstance(request.logger.handlers[0], listener)}  # type: ignore

    with create_test_client(route_handlers=[handler], logging_config=LoggingConfig(handlers=handlers)) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.json()["isinstance"]


def test_validation() -> None:
    logging_config = LoggingConfig(handlers={}, loggers={})
    assert logging_config.handlers["queue_listener"] == _get_default_handlers()["queue_listener"]
    assert logging_config.loggers["litestar"]


@pytest.mark.parametrize(
    "handlers, listener",
    [
        [default_handlers, StandardQueueListenerHandler],
        [default_picologging_handlers, PicologgingQueueListenerHandler],
    ],
)
def test_root_logger(handlers: Any, listener: Any) -> None:
    logging_config = LoggingConfig(handlers=handlers)
    get_logger = logging_config.configure()
    root_logger = get_logger()
    assert isinstance(root_logger.handlers[0], listener)  # type: ignore


@pytest.mark.parametrize(
    "handlers, listener",
    [
        pytest.param(
            default_handlers,
            StandardQueueListenerHandler,
            marks=pytest.mark.xfail(
                condition=sys.version_info >= (3, 12), reason="change to QueueHandler/QueueListener config in 3.12"
            ),
        ),
        [default_picologging_handlers, PicologgingQueueListenerHandler],
    ],
)
def test_customizing_handler(handlers: Any, listener: Any, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setitem(handlers["queue_listener"], "handlers", ["cfg://handlers.console"])

    logging_config = LoggingConfig(handlers=handlers)
    get_logger = logging_config.configure()
    root_logger = get_logger()
    assert isinstance(root_logger.handlers[0], listener)  # type: ignore



from docs.examples.middleware.session.cookies_full_example import app

from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT
from litestar.testing import TestClient


def test_session_middleware_example() -> None:
    with TestClient(app=app) as client:
        response = client.get("/session")
        assert response.status_code == HTTP_200_OK
        assert response.json() == {"has_session": False}

        response = client.post("/session")
        assert response.status_code == HTTP_201_CREATED

        response = client.get("/session")
        assert response.status_code == HTTP_200_OK
        assert response.json() == {"has_session": True}

        response = client.delete("/session")
        assert response.status_code == HTTP_204_NO_CONTENT

        response = client.get("/session")
        assert response.status_code == HTTP_200_OK
        assert response.json() == {"has_session": False}



from __future__ import annotations

import inspect
from functools import wraps
from typing import Callable, Literal, TypeVar
from warnings import warn

from typing_extensions import ParamSpec

__all__ = ("deprecated", "warn_deprecation")


T = TypeVar("T")
P = ParamSpec("P")
DeprecatedKind = Literal["function", "method", "classmethod", "attribute", "property", "class", "parameter", "import"]


def warn_deprecation(
    version: str,
    deprecated_name: str,
    kind: DeprecatedKind,
    *,
    removal_in: str | None = None,
    alternative: str | None = None,
    info: str | None = None,
    pending: bool = False,
) -> None:
    """Warn about a call to a (soon to be) deprecated function.

    Args:
        version: Litestar version where the deprecation will occur
        deprecated_name: Name of the deprecated function
        removal_in: Litestar version where the deprecated function will be removed
        alternative: Name of a function that should be used instead
        info: Additional information
        pending: Use ``PendingDeprecationWarning`` instead of ``DeprecationWarning``
        kind: Type of the deprecated thing
    """
    parts = []

    if kind == "import":
        access_type = "Import of"
    elif kind in {"function", "method"}:
        access_type = "Call to"
    else:
        access_type = "Use of"

    if pending:
        parts.append(f"{access_type} {kind} awaiting deprecation {deprecated_name!r}")
    else:
        parts.append(f"{access_type} deprecated {kind} {deprecated_name!r}")

    parts.extend(
        (
            f"Deprecated in litestar {version}",
            f"This {kind} will be removed in {removal_in or 'the next major version'}",
        )
    )
    if alternative:
        parts.append(f"Use {alternative!r} instead")

    if info:
        parts.append(info)

    text = ". ".join(parts)
    warning_class = PendingDeprecationWarning if pending else DeprecationWarning

    warn(text, warning_class, stacklevel=2)


def deprecated(
    version: str,
    *,
    removal_in: str | None = None,
    alternative: str | None = None,
    info: str | None = None,
    pending: bool = False,
    kind: Literal["function", "method", "classmethod", "property"] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Create a decorator wrapping a function, method or property with a warning call about a (pending) deprecation.

    Args:
        version: Litestar version where the deprecation will occur
        removal_in: Litestar version where the deprecated function will be removed
        alternative: Name of a function that should be used instead
        info: Additional information
        pending: Use ``PendingDeprecationWarning`` instead of ``DeprecationWarning``
        kind: Type of the deprecated callable. If ``None``, will use ``inspect`` to figure
            out if it's a function or method

    Returns:
        A decorator wrapping the function call with a warning
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
            warn_deprecation(
                version=version,
                deprecated_name=func.__name__,
                info=info,
                alternative=alternative,
                pending=pending,
                removal_in=removal_in,
                kind=kind or ("method" if inspect.ismethod(func) else "function"),
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator



from __future__ import annotations

from typing import Any

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sqlalchemy import Engine, StaticPool, create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from litestar.testing import TestClient


@pytest.fixture
def data() -> list[dict[str, Any]]:
    return [{"title": "test", "done": False}]


@pytest.fixture()
def sqlite_engine() -> Engine:
    return create_engine("sqlite://", connect_args={"check_same_thread": False}, poolclass=StaticPool)


@pytest.fixture()
def aiosqlite_engine() -> Engine:
    return create_async_engine("sqlite+aiosqlite://", connect_args={"check_same_thread": False}, poolclass=StaticPool)


def test_sqlalchemy_async_plugin_example(
    data: dict[str, Any], monkeypatch: MonkeyPatch, aiosqlite_engine: AsyncEngine
) -> None:
    from docs.examples.contrib.sqlalchemy.plugins import sqlalchemy_async_plugin_example

    monkeypatch.setattr(sqlalchemy_async_plugin_example.config, "engine_instance", aiosqlite_engine)

    with TestClient(sqlalchemy_async_plugin_example.app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_sync_plugin_example(data: dict[str, Any], monkeypatch: MonkeyPatch, sqlite_engine: Engine) -> None:
    from docs.examples.contrib.sqlalchemy.plugins import sqlalchemy_sync_plugin_example

    monkeypatch.setattr(sqlalchemy_sync_plugin_example.config, "engine_instance", sqlite_engine)

    with TestClient(sqlalchemy_sync_plugin_example.app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_async_init_plugin_example(
    data: dict[str, Any], monkeypatch: MonkeyPatch, aiosqlite_engine: AsyncEngine
) -> None:
    from docs.examples.contrib.sqlalchemy.plugins import sqlalchemy_async_init_plugin_example

    monkeypatch.setattr(sqlalchemy_async_init_plugin_example.config, "engine_instance", aiosqlite_engine)

    with TestClient(sqlalchemy_async_init_plugin_example.app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_sync_init_plugin_example(
    data: dict[str, Any], monkeypatch: MonkeyPatch, sqlite_engine: Engine
) -> None:
    from docs.examples.contrib.sqlalchemy.plugins import sqlalchemy_sync_init_plugin_example

    monkeypatch.setattr(sqlalchemy_sync_init_plugin_example.config, "engine_instance", sqlite_engine)

    with TestClient(sqlalchemy_sync_init_plugin_example.app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_async_init_plugin_dependencies(monkeypatch: MonkeyPatch) -> None:
    from docs.examples.contrib.sqlalchemy.plugins import sqlalchemy_async_dependencies

    monkeypatch.setattr(sqlalchemy_async_dependencies.config, "connection_string", "sqlite+aiosqlite://")

    with TestClient(sqlalchemy_async_dependencies.app) as client:
        assert client.post("/").json() == [1, 2]


def test_sqlalchemy_sync_init_plugin_dependencies(monkeypatch: MonkeyPatch) -> None:
    from docs.examples.contrib.sqlalchemy.plugins import sqlalchemy_sync_dependencies

    monkeypatch.setattr(sqlalchemy_sync_dependencies.config, "connection_string", "sqlite://")

    with TestClient(sqlalchemy_sync_dependencies.app) as client:
        assert client.post("/").json() == [1, 2]


def test_sqlalchemy_async_before_send_handler() -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_async_before_send_handler import app

    from litestar.contrib.sqlalchemy.plugins.init.config.asyncio import autocommit_before_send_handler

    assert autocommit_before_send_handler is app.before_send[0]


def test_sqlalchemy_sync_before_send_handler() -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_sync_before_send_handler import app

    from litestar.contrib.sqlalchemy.plugins.init.config.sync import autocommit_before_send_handler

    assert autocommit_before_send_handler is app.before_send[0].func


def test_sqlalchemy_async_serialization_plugin(data: dict[str, Any]) -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_async_serialization_plugin import app

    with TestClient(app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_sync_serialization_plugin(data: dict[str, Any]) -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_sync_serialization_plugin import app

    with TestClient(app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_async_serialization_dto(data: dict[str, Any]) -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_async_serialization_dto import app

    with TestClient(app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_async_serialization_plugin_marking_fields(data: dict[str, Any]) -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_async_serialization_plugin_marking_fields import app

    with TestClient(app) as client:
        assert client.post("/", json=data[0]).json() == data


def test_sqlalchemy_sync_serialization_plugin_marking_fields(data: dict[str, Any]) -> None:
    from docs.examples.contrib.sqlalchemy.plugins.sqlalchemy_sync_serialization_plugin_marking_fields import app

    with TestClient(app) as client:
        assert client.post("/", json=data[0]).json() == data



from datetime import datetime
from time import time
from typing import TYPE_CHECKING, Any

import pytest
from time_machine import travel

from litestar import Litestar, Request, get
from litestar.middleware.rate_limit import (
    DURATION_VALUES,
    CacheObject,
    DurationUnit,
    RateLimitConfig,
)
from litestar.serialization import decode_json, encode_json
from litestar.static_files.config import StaticFilesConfig
from litestar.status_codes import HTTP_200_OK, HTTP_429_TOO_MANY_REQUESTS
from litestar.stores.base import Store
from litestar.testing import TestClient, create_test_client

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize("unit", ["minute", "second", "hour", "day"])
async def test_rate_limiting(unit: DurationUnit) -> None:
    @get("/")
    def handler() -> None:
        return None

    config = RateLimitConfig(rate_limit=(unit, 1))
    cache_key = "RateLimitMiddleware::testclient"
    app = Litestar(route_handlers=[handler], middleware=[config.middleware])
    store = app.stores.get("rate_limit")

    with travel(datetime.utcnow, tick=False) as frozen_time, TestClient(app=app) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        cached_value = await store.get(cache_key)
        assert cached_value
        cache_object = CacheObject(**decode_json(value=cached_value))
        assert len(cache_object.history) == 1

        assert response.headers.get(config.rate_limit_policy_header_key) == f"1; w={DURATION_VALUES[unit]}"
        assert response.headers.get(config.rate_limit_limit_header_key) == "1"
        assert response.headers.get(config.rate_limit_remaining_header_key) == "0"
        assert response.headers.get(config.rate_limit_reset_header_key) == str(int(time()) - cache_object.reset)

        frozen_time.shift(DURATION_VALUES[unit] - 1)

        response = client.get("/")
        assert response.status_code == HTTP_429_TOO_MANY_REQUESTS
        assert response.headers.get(config.rate_limit_policy_header_key) == f"1; w={DURATION_VALUES[unit]}"
        assert response.headers.get(config.rate_limit_limit_header_key) == "1"
        assert response.headers.get(config.rate_limit_remaining_header_key) == "0"
        assert response.headers.get(config.rate_limit_reset_header_key) == str(int(time()) - cache_object.reset)

        frozen_time.shift(1)

        response = client.get("/")
        assert response.status_code == HTTP_200_OK


async def test_non_default_store(memory_store: Store) -> None:
    @get("/")
    def handler() -> None:
        return None

    app = Litestar(
        [handler], middleware=[RateLimitConfig(("second", 10)).middleware], stores={"rate_limit": memory_store}
    )

    with TestClient(app) as client:
        res = client.get("/")
        assert res.status_code == 200

    assert await memory_store.exists("RateLimitMiddleware::testclient")


async def test_set_store_name(memory_store: Store) -> None:
    @get("/")
    def handler() -> None:
        return None

    app = Litestar(
        [handler],
        middleware=[RateLimitConfig(("second", 10), store="some_store").middleware],
        stores={"some_store": memory_store},
    )

    with TestClient(app) as client:
        res = client.get("/")
        assert res.status_code == 200

    assert await memory_store.exists("RateLimitMiddleware::testclient")


async def test_reset() -> None:
    @get("/")
    def handler() -> None:
        return None

    config = RateLimitConfig(rate_limit=("second", 1))
    cache_key = "RateLimitMiddleware::testclient"
    app = Litestar(route_handlers=[handler], middleware=[config.middleware])
    store = app.stores.get("rate_limit")

    with TestClient(app=app) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        cached_value = await store.get(cache_key)
        assert cached_value
        cache_object = CacheObject(**decode_json(value=cached_value))
        assert cache_object.reset == int(time() + 1)

        cache_object.reset -= 2
        await store.set(cache_key, encode_json(cache_object))

        response = client.get("/")
        assert response.status_code == HTTP_200_OK


def test_exclude_patterns() -> None:
    @get("/excluded")
    def handler() -> None:
        return None

    @get("/not-excluded")
    def handler2() -> None:
        return None

    config = RateLimitConfig(rate_limit=("second", 1), exclude=["/excluded"])

    with create_test_client(route_handlers=[handler, handler2], middleware=[config.middleware]) as client:
        response = client.get("/excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/not-excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/not-excluded")
        assert response.status_code == HTTP_429_TOO_MANY_REQUESTS


def test_exclude_opt_key() -> None:
    @get("/excluded", skip_rate_limiting=True)
    def handler() -> None:
        return None

    @get("/not-excluded")
    def handler2() -> None:
        return None

    config = RateLimitConfig(rate_limit=("second", 1), exclude_opt_key="skip_rate_limiting")

    with create_test_client(route_handlers=[handler, handler2], middleware=[config.middleware]) as client:
        response = client.get("/excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/not-excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/not-excluded")
        assert response.status_code == HTTP_429_TOO_MANY_REQUESTS


def test_check_throttle_handler() -> None:
    @get("/path1")
    def handler1() -> None:
        return None

    @get("/path2")
    def handler2() -> None:
        return None

    def check_throttle_handler(request: Request[Any, Any, Any]) -> bool:
        return request.url.path == "/path1"

    config = RateLimitConfig(rate_limit=("minute", 1), check_throttle_handler=check_throttle_handler)

    with create_test_client(route_handlers=[handler1, handler2], middleware=[config.middleware]) as client:
        response = client.get("/path1")
        assert response.status_code == HTTP_200_OK

        response = client.get("/path1")
        assert response.status_code == HTTP_429_TOO_MANY_REQUESTS

        response = client.get("/path2")
        assert response.status_code == HTTP_200_OK

        response = client.get("/path2")
        assert response.status_code == HTTP_200_OK


async def test_rate_limiting_works_with_mounted_apps(tmpdir: "Path") -> None:
    # https://github.com/litestar-org/litestar/issues/781
    @get("/not-excluded")
    def handler() -> None:
        return None

    path1 = tmpdir / "test.css"
    path1.write_text("styles content", "utf-8")

    static_files_config = StaticFilesConfig(directories=[tmpdir], path="/src/static")  # pyright: ignore
    rate_limit_config = RateLimitConfig(rate_limit=("minute", 1), exclude=[r"^/src.*$"])
    with create_test_client(
        [handler], static_files_config=[static_files_config], middleware=[rate_limit_config.middleware]
    ) as client:
        response = client.get("/not-excluded")
        assert response.status_code == HTTP_200_OK

        response = client.get("/not-excluded")
        assert response.status_code == HTTP_429_TOO_MANY_REQUESTS

        response = client.get("/src/static/test.css")
        assert response.status_code == HTTP_200_OK
        assert response.text == "styles content"



from docs.examples.parameters.path_parameters_1 import app
from docs.examples.parameters.path_parameters_2 import app as app_2
from docs.examples.parameters.path_parameters_3 import app as app_3

from litestar.testing import TestClient


def test_path_parameters_1() -> None:
    with TestClient(app=app) as client:
        response = client.get("/user/1")
        assert response.status_code == 200
        assert response.json() == {"id": 1, "name": "John Doe"}


def test_path_parameters_2() -> None:
    with TestClient(app=app_2) as client:
        response = client.get("/orders/1667924386")
        assert response.status_code == 200
        assert response.json() == [
            {"id": 1, "customer_id": 2},
            {"id": 2, "customer_id": 2},
        ]


def test_path_parameters_3() -> None:
    with TestClient(app=app_3) as client:
        response = client.get("/versions/1")
        assert response.status_code == 200
        assert response.json() == {"id": 1, "specs": {"some": "value"}}



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Union

if TYPE_CHECKING:
    from litestar.openapi.spec.reference import Reference
    from litestar.openapi.spec.response import OpenAPIResponse

Responses = Dict[str, Union["OpenAPIResponse", "Reference"]]
"""A container for the expected responses of an operation. The container maps a
HTTP response code to the expected response.

The documentation is not necessarily expected to cover all possible HTTP response codes because they may not be known in
advance. However, documentation is expected to cover a successful operation response and any known errors.

The ``default`` MAY be used as a default response object for all HTTP codes hat are not covered individually by the
specification.

The ``Responses Object`` MUST contain at least one response code, and it SHOULD be the response for a successful
operation call.

Fixed Fields

default: ``Optional[Union[Response, Reference]]``

The documentation of responses other than the ones declared for specific HTTP response codes. Use this field to cover
undeclared responses. A `Reference Object <https://spec.openapis.org/oas/v3.1.0#referenceObject>`_ can link to a
response that the `OpenAPI Object's components/responses <https://spec.openapis.org/oas/v3.1.0#componentsResponses>`_
section defines.

Patterned Fields
{httpStatusCode}: ``Optional[Union[Response, Reference]]``

Any `HTTP status code <https://spec.openapis.org/oas/v3.1.0#httpCodes>`_ can be used as the property name, but only one
property per code, to describe the expected response for that HTTP status code.

A `Reference Object <https://spec.openapis.org/oas/v3.1.0#referenceObject>`_ can link to a response that is defined in
the `OpenAPI Object's components/responses <https://spec.openapis.org/oas/v3.1.0#componentsResponses>`_ section. This
field MUST be enclosed in quotation marks (for example, ``200``) for compatibility between JSON and YAML. To define a
range of response codes, this field MAY contain the uppercase wildcard character ``X``. For example, ``2XX`` represents
all response codes between ``[200-299]``. Only the following range definitions are allowed: ``1XX``, ``2XX``, ``3XX``,
``4XX``, and ``5XX``. If a response is defined using an explicit code, the explicit code definition takes precedence
over the range definition for that code.
"""



from typing import Awaitable, Callable

from litestar import Controller, Litestar, Router, get
from litestar.di import Provide


def test_resolve_dependencies_without_provide() -> None:
    async def foo() -> None:
        pass

    async def bar() -> None:
        pass

    @get(dependencies={"foo": foo, "bar": Provide(bar)})
    async def handler() -> None:
        pass

    assert handler.resolve_dependencies() == {"foo": Provide(foo), "bar": Provide(bar)}


def function_factory() -> Callable[[], Awaitable[None]]:
    async def func() -> None:
        return None

    return func


def test_resolve_from_layers() -> None:
    app_dependency = function_factory()
    router_dependency = function_factory()
    controller_dependency = function_factory()
    handler_dependency = function_factory()

    class MyController(Controller):
        path = "/controller"
        dependencies = {"controller": controller_dependency}

        @get("/handler", dependencies={"handler": handler_dependency}, name="foo")
        async def handler(self) -> None:
            pass

    router = Router("/router", route_handlers=[MyController], dependencies={"router": router_dependency})
    app = Litestar([router], dependencies={"app": app_dependency}, openapi_config=None)

    handler_map = app.get_handler_index_by_name("foo")
    assert handler_map
    handler = handler_map["handler"]

    assert handler.resolve_dependencies() == {
        "app": Provide(app_dependency),
        "router": Provide(router_dependency),
        "controller": Provide(controller_dependency),
        "handler": Provide(handler_dependency),
    }



import logging
from typing import TYPE_CHECKING

import pytest
from docs.examples.responses.background_tasks_1 import app as app_1
from docs.examples.responses.background_tasks_2 import app as app_2
from docs.examples.responses.background_tasks_3 import app as app_3
from docs.examples.responses.background_tasks_3 import greeted as greeted_3

from litestar.testing import TestClient

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


pytestmark = pytest.mark.usefixtures("reset_httpx_logging")


def test_background_tasks_1(caplog: "LogCaptureFixture") -> None:
    with caplog.at_level(logging.INFO), TestClient(app=app_1) as client:
        name = "Jane"
        res = client.get("/", params={"name": name})
        assert res.status_code == 200
        assert res.json()["hello"] == name
        assert len(caplog.messages) == 1
        assert name in caplog.messages[0]


def test_background_tasks_2(caplog: "LogCaptureFixture") -> None:
    with caplog.at_level(logging.INFO), TestClient(app=app_2) as client:
        res = client.get("/")
        assert res.status_code == 200
        assert "hello" in res.json()
        assert len(caplog.messages) == 1
        assert "greeter" in caplog.messages[0]


def test_background_tasks_3(caplog: "LogCaptureFixture") -> None:
    with caplog.at_level(logging.INFO), TestClient(app=app_3) as client:
        name = "Jane"
        res = client.get("/", params={"name": name})
        assert res.status_code == 200
        assert res.json()["hello"] == name
        assert len(caplog.messages) == 1
        assert name in caplog.messages[0]
        assert name in greeted_3



from typing import Any, Dict, Generic, Tuple, TypeVar

import pytest
from pydantic import BaseModel

from litestar.contrib.pydantic.utils import pydantic_get_type_hints_with_generics_resolved

T = TypeVar("T")


class GenericPydanticModel(BaseModel, Generic[T]):
    foo: T


@pytest.mark.parametrize(
    ("annotation", "expected_type_hints"),
    (
        (GenericPydanticModel, {"foo": T}),
        (GenericPydanticModel[int], {"foo": int}),
    ),
)
def test_get_pydantic_type_hints_with_generics_resolved(annotation: Any, expected_type_hints: Dict[str, Any]) -> None:
    type_hints = pydantic_get_type_hints_with_generics_resolved(annotation)

    # In Python 3.12 and Pydantic V1, `__slots__` is returned in `get_type_hints`.
    slots_type = type_hints.pop("__slots__", None)
    if slots_type is not None:
        assert slots_type == Tuple[str, ...]

    assert type_hints == expected_type_hints



from typing import Any

from litestar import Litestar, get
from litestar.connection import ASGIConnection
from litestar.logging.config import LoggingConfig
from litestar.testing import RequestFactory
from litestar.types.empty import Empty
from litestar.utils.scope.state import ScopeState


def test_connection_base_properties() -> None:
    @get("/")
    def handler() -> None:
        return None

    app = Litestar(route_handlers=[handler], logging_config=LoggingConfig())
    user = {"name": "moishe"}
    auth = {"key": "value"}
    session = {"session": "abc"}
    scope = RequestFactory(app=app).get(route_handler=handler, user=user, auth=auth, session=session).scope
    connection = ASGIConnection[Any, Any, Any, Any](scope)
    connection_state = ScopeState.from_scope(scope)

    assert connection.app
    assert connection.app is app
    assert connection.route_handler is handler
    assert connection.state is not None
    assert connection_state.url is Empty
    assert connection.url
    assert connection_state.url is not Empty
    assert connection_state.base_url is Empty  # type:ignore[unreachable]
    assert connection.base_url
    assert connection_state.base_url is not Empty
    assert connection_state.headers is Empty
    assert connection.headers is not None
    assert connection_state.headers is not Empty
    assert connection_state.parsed_query is Empty
    assert connection.query_params is not None
    assert connection_state.parsed_query is not Empty
    assert connection_state.cookies is Empty
    assert connection.cookies is not None
    assert connection_state.cookies is not Empty
    assert connection.client
    assert connection.user is user
    assert connection.auth is auth
    assert connection.session is session



"""Example domain objects for testing."""
from __future__ import annotations

from datetime import date, datetime
from typing import List

from advanced_alchemy import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.base import BigIntAuditBase, BigIntBase
from sqlalchemy import Column, FetchedValue, ForeignKey, String, Table, func
from sqlalchemy.orm import Mapped, mapped_column, relationship


class BigIntAuthor(BigIntAuditBase):
    """The Author domain object."""

    name: Mapped[str] = mapped_column(String(length=100))  # pyright: ignore
    dob: Mapped[date] = mapped_column(nullable=True)  # pyright: ignore
    books: Mapped[List[BigIntBook]] = relationship(  # noqa
        lazy="selectin",
        back_populates="author",
        cascade="all, delete",
    )


class BigIntBook(BigIntBase):
    """The Book domain object."""

    title: Mapped[str] = mapped_column(String(length=250))  # pyright: ignore
    author_id: Mapped[int] = mapped_column(ForeignKey("big_int_author.id"))  # pyright: ignore
    author: Mapped[BigIntAuthor] = relationship(  # pyright: ignore
        lazy="joined",
        innerjoin=True,
        back_populates="books",
    )


class BigIntEventLog(BigIntAuditBase):
    """The event log domain object."""

    logged_at: Mapped[datetime] = mapped_column(default=datetime.now())  # pyright: ignore
    payload: Mapped[dict] = mapped_column(default=lambda: {})  # pyright: ignore


class BigIntModelWithFetchedValue(BigIntBase):
    """The ModelWithFetchedValue BigIntBase."""

    val: Mapped[int]  # pyright: ignore
    updated: Mapped[datetime] = mapped_column(  # pyright: ignore
        server_default=func.current_timestamp(),
        onupdate=func.current_timestamp(),
        server_onupdate=FetchedValue(),
    )


bigint_item_tag = Table(
    "bigint_item_tag",
    BigIntBase.metadata,
    Column("item_id", ForeignKey("big_int_item.id"), primary_key=True),
    Column("tag_id", ForeignKey("big_int_tag.id"), primary_key=True),
)


class BigIntItem(BigIntBase):
    name: Mapped[str] = mapped_column(String(length=50))  # pyright: ignore
    description: Mapped[str] = mapped_column(String(length=100), nullable=True)  # pyright: ignore
    tags: Mapped[List[BigIntTag]] = relationship(secondary=lambda: bigint_item_tag, back_populates="items")  # noqa


class BigIntTag(BigIntBase):
    """The event log domain object."""

    name: Mapped[str] = mapped_column(String(length=50))  # pyright: ignore
    items: Mapped[List[BigIntItem]] = relationship(secondary=lambda: bigint_item_tag, back_populates="tags")  # noqa


class BigIntRule(BigIntAuditBase):
    """The rule domain object."""

    name: Mapped[str] = mapped_column(String(length=250))  # pyright: ignore
    config: Mapped[dict] = mapped_column(default=lambda: {})  # pyright: ignore


class RuleAsyncRepository(SQLAlchemyAsyncRepository[BigIntRule]):
    """Rule repository."""

    model_type = BigIntRule


class AuthorAsyncRepository(SQLAlchemyAsyncRepository[BigIntAuthor]):
    """Author repository."""

    model_type = BigIntAuthor


class BookAsyncRepository(SQLAlchemyAsyncRepository[BigIntBook]):
    """Book repository."""

    model_type = BigIntBook


class EventLogAsyncRepository(SQLAlchemyAsyncRepository[BigIntEventLog]):
    """Event log repository."""

    model_type = BigIntEventLog


class ModelWithFetchedValueAsyncRepository(SQLAlchemyAsyncRepository[BigIntModelWithFetchedValue]):
    """BigIntModelWithFetchedValue repository."""

    model_type = BigIntModelWithFetchedValue


class TagAsyncRepository(SQLAlchemyAsyncRepository[BigIntTag]):
    """Tag repository."""

    model_type = BigIntTag


class ItemAsyncRepository(SQLAlchemyAsyncRepository[BigIntItem]):
    """Item repository."""

    model_type = BigIntItem


class AuthorSyncRepository(SQLAlchemySyncRepository[BigIntAuthor]):
    """Author repository."""

    model_type = BigIntAuthor


class BookSyncRepository(SQLAlchemySyncRepository[BigIntBook]):
    """Book repository."""

    model_type = BigIntBook


class EventLogSyncRepository(SQLAlchemySyncRepository[BigIntEventLog]):
    """Event log repository."""

    model_type = BigIntEventLog


class RuleSyncRepository(SQLAlchemySyncRepository[BigIntRule]):
    """Rule repository."""

    model_type = BigIntRule


class ModelWithFetchedValueSyncRepository(SQLAlchemySyncRepository[BigIntModelWithFetchedValue]):
    """ModelWithFetchedValue repository."""

    model_type = BigIntModelWithFetchedValue


class TagSyncRepository(SQLAlchemySyncRepository[BigIntTag]):
    """Tag repository."""

    model_type = BigIntTag


class ItemSyncRepository(SQLAlchemySyncRepository[BigIntItem]):
    """Item repository."""

    model_type = BigIntItem



from __future__ import annotations

import os
import shutil
import unicodedata
from tempfile import mkstemp
from typing import TYPE_CHECKING

from anyio import Path
from anyio.to_thread import run_sync

from .base import NamespacedStore, StorageObject

__all__ = ("FileStore",)


if TYPE_CHECKING:
    from datetime import timedelta
    from os import PathLike


def _safe_file_name(name: str) -> str:
    name = unicodedata.normalize("NFKD", name)
    return "".join(c if c.isalnum() else str(ord(c)) for c in name)


class FileStore(NamespacedStore):
    """File based, thread and process safe, asynchronous key/value store."""

    __slots__ = {"path": "file path"}

    def __init__(self, path: PathLike[str]) -> None:
        """Initialize ``FileStorage``.

        Args:
            path: Path to store data under
        """
        self.path = Path(path)

    def with_namespace(self, namespace: str) -> FileStore:
        """Return a new instance of :class:`FileStore`, using  a sub-path of the current store's path."""
        if not namespace.isalnum():
            raise ValueError(f"Invalid namespace: {namespace!r}")
        return FileStore(self.path / namespace)

    def _path_from_key(self, key: str) -> Path:
        return self.path / _safe_file_name(key)

    @staticmethod
    async def _load_from_path(path: Path) -> StorageObject | None:
        try:
            data = await path.read_bytes()
            return StorageObject.from_bytes(data)
        except FileNotFoundError:
            return None

    def _write_sync(self, target_file: Path, storage_obj: StorageObject) -> None:
        try:
            tmp_file_fd, tmp_file_name = mkstemp(dir=self.path, prefix=f"{target_file.name}.tmp")
            renamed = False
            try:
                try:
                    os.write(tmp_file_fd, storage_obj.to_bytes())
                finally:
                    os.close(tmp_file_fd)

                os.replace(tmp_file_name, target_file)  # noqa: PTH105
                renamed = True
            finally:
                if not renamed:
                    os.unlink(tmp_file_name)  # noqa: PTH108
        except OSError:
            pass

    async def _write(self, target_file: Path, storage_obj: StorageObject) -> None:
        await run_sync(self._write_sync, target_file, storage_obj)

    async def set(self, key: str, value: str | bytes, expires_in: int | timedelta | None = None) -> None:
        """Set a value.

        Args:
            key: Key to associate the value with
            value: Value to store
            expires_in: Time in seconds before the key is considered expired

        Returns:
            ``None``
        """

        await self.path.mkdir(exist_ok=True)
        path = self._path_from_key(key)
        if isinstance(value, str):
            value = value.encode("utf-8")
        storage_obj = StorageObject.new(data=value, expires_in=expires_in)
        await self._write(path, storage_obj)

    async def get(self, key: str, renew_for: int | timedelta | None = None) -> bytes | None:
        """Get a value.

        Args:
            key: Key associated with the value
            renew_for: If given and the value had an initial expiry time set, renew the
                expiry time for ``renew_for`` seconds. If the value has not been set
                with an expiry time this is a no-op

        Returns:
            The value associated with ``key`` if it exists and is not expired, else
            ``None``
        """
        path = self._path_from_key(key)
        storage_obj = await self._load_from_path(path)

        if not storage_obj:
            return None

        if storage_obj.expired:
            await path.unlink(missing_ok=True)
            return None

        if renew_for and storage_obj.expires_at:
            await self.set(key, value=storage_obj.data, expires_in=renew_for)

        return storage_obj.data

    async def delete(self, key: str) -> None:
        """Delete a value.

        If no such key exists, this is a no-op.

        Args:
            key: Key of the value to delete
        """
        path = self._path_from_key(key)
        await path.unlink(missing_ok=True)

    async def delete_all(self) -> None:
        """Delete all stored values.

        Note:
            This deletes and recreates :attr:`FileStore.path`
        """

        await run_sync(shutil.rmtree, self.path)
        await self.path.mkdir(exist_ok=True)

    async def delete_expired(self) -> None:
        """Delete expired items.

        Since expired items are normally only cleared on access (i.e. when calling
        :meth:`.get`), this method should be called in regular intervals
        to free disk space.
        """
        async for file in self.path.iterdir():
            wrapper = await self._load_from_path(file)
            if wrapper and wrapper.expired:
                await file.unlink(missing_ok=True)

    async def exists(self, key: str) -> bool:
        """Check if a given ``key`` exists."""
        path = self._path_from_key(key)
        return await path.exists()

    async def expires_in(self, key: str) -> int | None:
        """Get the time in seconds ``key`` expires in. If no such ``key`` exists or no
        expiry time was set, return ``None``.
        """
        if storage_obj := await self._load_from_path(self._path_from_key(key)):
            return storage_obj.expires_in
        return None



from __future__ import annotations

from dataclasses import dataclass
from os import environ
from typing import Any
from uuid import UUID

from litestar import Request, Response, get, post
from litestar.connection import ASGIConnection
from litestar.security.jwt import JWTAuth, Token
from litestar.status_codes import HTTP_204_NO_CONTENT
from litestar.testing import create_test_client


@dataclass
class User:
    id: UUID
    name: str
    email: str


MOCK_DB: dict[str, User] = {}


async def retrieve_user_handler(token: Token, connection: ASGIConnection[Any, Any, Any, Any]) -> User | None:
    return MOCK_DB.get(token.sub)


jwt_auth = JWTAuth[User](
    retrieve_user_handler=retrieve_user_handler,
    token_secret=environ.get("JWT_SECRET", "abcd123"),
    exclude=["/login", "/schema"],
)


@post("/login")
async def login_handler(data: User) -> Response[User]:
    MOCK_DB[str(data.id)] = data
    return jwt_auth.login(identifier=str(data.id), response_body=data)


@get("/some-path", sync_to_thread=False)
def some_route_handler(request: Request[User, Token, Any]) -> Any:
    assert isinstance(request.user, User)


def test_options_request_with_jwt() -> None:
    with create_test_client(
        route_handlers=[login_handler, some_route_handler],
        on_app_init=[jwt_auth.on_app_init],
    ) as client:
        response = client.options("/some-path")
        assert response.status_code == HTTP_204_NO_CONTENT



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from litestar.enums import MediaType

if TYPE_CHECKING:
    from litestar.types import DataContainerType


__all__ = ("ResponseSpec",)


@dataclass
class ResponseSpec:
    """Container type of additional responses."""

    data_container: DataContainerType
    """A model that describes the content of the response."""
    generate_examples: bool = field(default=True)
    """Generate examples for the response content."""
    description: str = field(default="Additional response")
    """A description of the response."""
    media_type: MediaType = field(default=MediaType.JSON)
    """Response media type."""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Sequence

from litestar.exceptions import ImproperlyConfiguredException
from litestar.handlers.base import BaseRouteHandler
from litestar.types.builtin_types import NoneType
from litestar.utils.predicates import is_async_callable

__all__ = ("ASGIRouteHandler", "asgi")


if TYPE_CHECKING:
    from litestar.types import (
        ExceptionHandlersMap,
        Guard,
        MaybePartial,  # noqa: F401
    )


class ASGIRouteHandler(BaseRouteHandler):
    """ASGI Route Handler decorator.

    Use this decorator to decorate ASGI applications.
    """

    __slots__ = ("is_mount", "is_static")

    def __init__(
        self,
        path: str | Sequence[str] | None = None,
        *,
        exception_handlers: ExceptionHandlersMap | None = None,
        guards: Sequence[Guard] | None = None,
        name: str | None = None,
        opt: Mapping[str, Any] | None = None,
        is_mount: bool = False,
        is_static: bool = False,
        signature_namespace: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize ``ASGIRouteHandler``.

        Args:
            exception_handlers: A mapping of status codes and/or exception types to handler functions.
            guards: A sequence of :class:`Guard <.types.Guard>` callables.
            name: A string identifying the route handler.
            opt: A string key mapping of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or
                wherever you have access to :class:`Request <.connection.Request>` or
                :class:`ASGI Scope <.types.Scope>`.
            path: A path fragment for the route handler function or a list of path fragments. If not given defaults to
                ``/``
            is_mount: A boolean dictating whether the handler's paths should be regarded as mount paths. Mount path
                accept any arbitrary paths that begin with the defined prefixed path. For example, a mount with the path
                ``/some-path/`` will accept requests for ``/some-path/`` and any sub path under this, e.g.
                ``/some-path/sub-path/`` etc.
            is_static: A boolean dictating whether the handler's paths should be regarded as static paths. Static paths
                are used to deliver static files.
            signature_namespace: A mapping of names to types for use in forward reference resolution during signature modelling.
            type_encoders: A mapping of types to callables that transform them into types supported for serialization.
            **kwargs: Any additional kwarg - will be set in the opt dictionary.
        """
        self.is_mount = is_mount or is_static
        self.is_static = is_static
        super().__init__(
            path,
            exception_handlers=exception_handlers,
            guards=guards,
            name=name,
            opt=opt,
            signature_namespace=signature_namespace,
            **kwargs,
        )

    def _validate_handler_function(self) -> None:
        """Validate the route handler function once it's set by inspecting its return annotations."""
        super()._validate_handler_function()

        if not self.parsed_fn_signature.return_type.is_subclass_of(NoneType):
            raise ImproperlyConfiguredException("ASGI handler functions should return 'None'")

        if any(key not in self.parsed_fn_signature.parameters for key in ("scope", "send", "receive")):
            raise ImproperlyConfiguredException(
                "ASGI handler functions should define 'scope', 'send' and 'receive' arguments"
            )
        if not is_async_callable(self.fn):
            raise ImproperlyConfiguredException("Functions decorated with 'asgi' must be async functions")


asgi = ASGIRouteHandler



from __future__ import annotations

import enum
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Sequence

from litestar.config.allowed_hosts import AllowedHostsConfig
from litestar.config.response_cache import ResponseCacheConfig
from litestar.datastructures import State
from litestar.events.emitter import SimpleEventEmitter
from litestar.types.empty import Empty

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from litestar import Litestar
    from litestar.config.compression import CompressionConfig
    from litestar.config.cors import CORSConfig
    from litestar.config.csrf import CSRFConfig
    from litestar.connection import Request, WebSocket
    from litestar.datastructures import CacheControlHeader, ETag, ResponseHeader
    from litestar.di import Provide
    from litestar.dto import AbstractDTO
    from litestar.events.emitter import BaseEventEmitterBackend
    from litestar.events.listener import EventListener
    from litestar.logging.config import BaseLoggingConfig
    from litestar.openapi.config import OpenAPIConfig
    from litestar.openapi.spec import SecurityRequirement
    from litestar.plugins import PluginProtocol
    from litestar.static_files.config import StaticFilesConfig
    from litestar.stores.base import Store
    from litestar.stores.registry import StoreRegistry
    from litestar.types import (
        AfterExceptionHookHandler,
        AfterRequestHookHandler,
        AfterResponseHookHandler,
        AnyCallable,
        BeforeMessageSendHookHandler,
        BeforeRequestHookHandler,
        ControllerRouterHandler,
        ExceptionHandlersMap,
        Guard,
        Middleware,
        ParametersMap,
        ResponseCookies,
        ResponseType,
        TypeEncodersMap,
    )
    from litestar.types.callable_types import LifespanHook
    from litestar.types.composite_types import TypeDecodersSequence
    from litestar.types.empty import EmptyType
    from litestar.types.internal_types import TemplateConfigType


__all__ = (
    "AppConfig",
    "ExperimentalFeatures",
)


@dataclass
class AppConfig:
    """The parameters provided to the ``Litestar`` app are used to instantiate an instance, and then the instance is
    passed to any callbacks registered to ``on_app_init`` in the order they are provided.

    The final attribute values are used to instantiate the application object.
    """

    after_exception: list[AfterExceptionHookHandler] = field(default_factory=list)
    """An application level :class:`exception hook handler <.types.AfterExceptionHookHandler>` or list thereof.

    This hook is called after an exception occurs. In difference to exception handlers, it is not meant to return a
    response - only to process the exception (e.g. log it, send it to Sentry etc.).
    """
    after_request: AfterRequestHookHandler | None = field(default=None)
    """A sync or async function executed after the route handler function returned and the response object has been
    resolved.

    Receives the response object which may be any subclass of :class:`Response <.response.Response>`.
    """
    after_response: AfterResponseHookHandler | None = field(default=None)
    """A sync or async function called after the response has been awaited. It receives the
    :class:`Request <.connection.Request>` object and should not return any values.
    """
    allowed_hosts: list[str] | AllowedHostsConfig | None = field(default=None)
    """If set enables the builtin allowed hosts middleware."""
    before_request: BeforeRequestHookHandler | None = field(default=None)
    """A sync or async function called immediately before calling the route handler. Receives the
    :class:`Request <.connection.Request>` instance and any non-``None`` return value is used for the response,
    bypassing the route handler.
    """
    before_send: list[BeforeMessageSendHookHandler] = field(default_factory=list)
    """An application level :class:`before send hook handler <.types.BeforeMessageSendHookHandler>` or list thereof.

    This hook is called when the ASGI send function is called.
    """
    cache_control: CacheControlHeader | None = field(default=None)
    """A ``cache-control`` header of type :class:`CacheControlHeader <.datastructures.CacheControlHeader>` to add to
    route handlers of this app.

    Can be overridden by route handlers.
    """
    compression_config: CompressionConfig | None = field(default=None)
    """Configures compression behaviour of the application, this enabled a builtin or user defined Compression
    middleware.
    """
    cors_config: CORSConfig | None = field(default=None)
    """If set this enables the builtin CORS middleware."""
    csrf_config: CSRFConfig | None = field(default=None)
    """If set this enables the builtin CSRF middleware."""
    debug: bool = field(default=False)
    """If ``True``, app errors rendered as HTML with a stack trace."""
    dependencies: dict[str, Provide | AnyCallable] = field(default_factory=dict)
    """A string keyed dictionary of dependency :class:`Provider <.di.Provide>` instances."""
    dto: type[AbstractDTO] | None | EmptyType = field(default=Empty)
    """:class:`AbstractDTO <.dto.base_dto.AbstractDTO>` to use for (de)serializing and validation of request data."""
    etag: ETag | None = field(default=None)
    """An ``etag`` header of type :class:`ETag <.datastructures.ETag>` to add to route handlers of this app.

    Can be overridden by route handlers.
    """
    event_emitter_backend: type[BaseEventEmitterBackend] = field(default=SimpleEventEmitter)
    """A subclass of :class:`BaseEventEmitterBackend <.events.emitter.BaseEventEmitterBackend>`."""
    exception_handlers: ExceptionHandlersMap = field(default_factory=dict)
    """A dictionary that maps handler functions to status codes and/or exception types."""
    guards: list[Guard] = field(default_factory=list)
    """A list of :class:`Guard <.types.Guard>` callables."""
    include_in_schema: bool | EmptyType = field(default=Empty)
    """A boolean flag dictating whether  the route handler should be documented in the OpenAPI schema"""
    lifespan: list[Callable[[Litestar], AbstractAsyncContextManager] | AbstractAsyncContextManager] = field(
        default_factory=list
    )
    """A list of callables returning async context managers, wrapping the lifespan of the ASGI application"""
    listeners: list[EventListener] = field(default_factory=list)
    """A list of :class:`EventListener <.events.listener.EventListener>`."""
    logging_config: BaseLoggingConfig | None = field(default=None)
    """An instance of :class:`BaseLoggingConfig <.logging.config.BaseLoggingConfig>` subclass."""
    middleware: list[Middleware] = field(default_factory=list)
    """A list of :class:`Middleware <.types.Middleware>`."""
    on_shutdown: list[LifespanHook] = field(default_factory=list)
    """A list of :class:`LifespanHook <.types.LifespanHook>` called during application shutdown."""
    on_startup: list[LifespanHook] = field(default_factory=list)
    """A list of :class:`LifespanHook <.types.LifespanHook>` called during application startup."""
    openapi_config: OpenAPIConfig | None = field(default=None)
    """Defaults to :data:`DEFAULT_OPENAPI_CONFIG <litestar.app.DEFAULT_OPENAPI_CONFIG>`"""
    opt: dict[str, Any] = field(default_factory=dict)
    """A string keyed dictionary of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or
    wherever you have access to :class:`Request <.connection.Request>` or :class:`ASGI Scope <litestar.types.Scope>`.

    Can be overridden by routers and router handlers.
    """
    parameters: ParametersMap = field(default_factory=dict)
    """A mapping of :class:`Parameter <.params.Parameter>` definitions available to all application paths."""
    pdb_on_exception: bool = field(default=False)
    """Drop into the PDB on an exception"""
    plugins: list[PluginProtocol] = field(default_factory=list)
    """List of :class:`SerializationPluginProtocol <.plugins.SerializationPluginProtocol>`."""
    request_class: type[Request] | None = field(default=None)
    """An optional subclass of :class:`Request <.connection.Request>` to use for http connections."""
    response_class: ResponseType | None = field(default=None)
    """A custom subclass of :class:`Response <.response.Response>` to be used as the app's default response."""
    response_cookies: ResponseCookies = field(default_factory=list)
    """A list of :class:`Cookie <.datastructures.Cookie>`."""
    response_headers: Sequence[ResponseHeader] = field(default_factory=list)
    """A string keyed dictionary mapping :class:`ResponseHeader <.datastructures.ResponseHeader>`."""
    response_cache_config: ResponseCacheConfig = field(default_factory=ResponseCacheConfig)
    """Configures caching behavior of the application."""
    return_dto: type[AbstractDTO] | None | EmptyType = field(default=Empty)
    """:class:`AbstractDTO <.dto.base_dto.AbstractDTO>` to use for serializing outbound response
    data.
    """
    route_handlers: list[ControllerRouterHandler] = field(default_factory=list)
    """A required list of route handlers, which can include instances of :class:`Router <.router.Router>`,
    subclasses of :class:`Controller <.controller.Controller>` or any function decorated by the route handler
    decorators.
    """
    security: list[SecurityRequirement] = field(default_factory=list)
    """A list of dictionaries that will be added to the schema of all route handlers in the application. See
    :data:`SecurityRequirement <.openapi.spec.SecurityRequirement>` for details.
    """
    signature_namespace: dict[str, Any] = field(default_factory=dict)
    """A mapping of names to types for use in forward reference resolution during signature modeling."""
    signature_types: list[Any] = field(default_factory=list)
    """A sequence of types for use in forward reference resolution during signature modeling.

    These types will be added to the signature namespace using their ``__name__`` attribute.
    """
    state: State = field(default_factory=State)
    """A :class:`State` <.datastructures.State>` instance holding application state."""
    static_files_config: list[StaticFilesConfig] = field(default_factory=list)
    """An instance or list of :class:`StaticFilesConfig <.static_files.StaticFilesConfig>`."""
    stores: StoreRegistry | dict[str, Store] | None = None
    """Central registry of :class:`Store <.stores.base.Store>` to be made available and be used throughout the
    application. Can be either a dictionary mapping strings to :class:`Store <.stores.base.Store>` instances, or an
    instance of :class:`StoreRegistry <.stores.registry.StoreRegistry>`.
    """
    tags: list[str] = field(default_factory=list)
    """A list of string tags that will be appended to the schema of all route handlers under the application."""
    template_config: TemplateConfigType | None = field(default=None)
    """An instance of :class:`TemplateConfig <.template.TemplateConfig>`."""
    type_encoders: TypeEncodersMap | None = field(default=None)
    """A mapping of types to callables that transform them into types supported for serialization."""
    type_decoders: TypeDecodersSequence | None = field(default=None)
    """A sequence of tuples, each composed of a predicate testing for type identity and a msgspec hook for deserialization."""
    websocket_class: type[WebSocket] | None = field(default=None)
    """An optional subclass of :class:`WebSocket <.connection.WebSocket>` to use for websocket connections."""
    multipart_form_part_limit: int = field(default=1000)
    """The maximal number of allowed parts in a multipart/formdata request. This limit is intended to protect from
    DoS attacks."""
    experimental_features: list[ExperimentalFeatures] | None = None

    def __post_init__(self) -> None:
        """Normalize the allowed hosts to be a config or None.

        Returns:
            Optional config.
        """
        if self.allowed_hosts and isinstance(self.allowed_hosts, list):
            self.allowed_hosts = AllowedHostsConfig(allowed_hosts=self.allowed_hosts)


class ExperimentalFeatures(str, enum.Enum):
    DTO_CODEGEN = "DTO_CODEGEN"



from __future__ import annotations

import re
from collections import abc, defaultdict, deque
from typing import (
    AbstractSet,
    Any,
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Collection,
    Container,
    Coroutine,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Generator,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    MappingView,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Reversible,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    ValuesView,
    cast,
)

from typing_extensions import Annotated, NotRequired, Required, get_args, get_origin, get_type_hints

from litestar.types.builtin_types import NoneType, UnionTypes

__all__ = (
    "get_instantiable_origin",
    "get_origin_or_inner_type",
    "get_safe_generic_origin",
    "instantiable_type_mapping",
    "make_non_optional_union",
    "safe_generic_origin_map",
    "unwrap_annotation",
)


T = TypeVar("T")
UnionT = TypeVar("UnionT", bound="Union")

tuple_types_regex = re.compile(
    "^"
    + "|".join(
        [*[repr(x) for x in (List, Sequence, Iterable, Iterator, Tuple, Deque)], "tuple", "list", "collections.deque"]
    )
)

instantiable_type_mapping = {
    AbstractSet: set,
    DefaultDict: defaultdict,
    Deque: deque,
    Dict: dict,
    FrozenSet: frozenset,
    List: list,
    Mapping: dict,
    MutableMapping: dict,
    MutableSequence: list,
    MutableSet: set,
    Sequence: list,
    Set: set,
    Tuple: tuple,
    abc.Mapping: dict,
    abc.MutableMapping: dict,
    abc.MutableSequence: list,
    abc.MutableSet: set,
    abc.Sequence: list,
    abc.Set: set,
    defaultdict: defaultdict,
    deque: deque,
    dict: dict,
    frozenset: frozenset,
    list: list,
    set: set,
    tuple: tuple,
}

safe_generic_origin_map = {
    set: AbstractSet,
    defaultdict: DefaultDict,
    deque: Deque,
    dict: Dict,
    frozenset: FrozenSet,
    list: List,
    tuple: Tuple,
    abc.Mapping: Mapping,
    abc.MutableMapping: MutableMapping,
    abc.MutableSequence: MutableSequence,
    abc.MutableSet: MutableSet,
    abc.Sequence: Sequence,
    abc.Set: AbstractSet,
    abc.Collection: Collection,
    abc.Container: Container,
    abc.ItemsView: ItemsView,
    abc.KeysView: KeysView,
    abc.MappingView: MappingView,
    abc.ValuesView: ValuesView,
    abc.Iterable: Iterable,
    abc.Iterator: Iterator,
    abc.Generator: Generator,
    abc.Reversible: Reversible,
    abc.Coroutine: Coroutine,
    abc.AsyncGenerator: AsyncGenerator,
    abc.AsyncIterable: AsyncIterable,
    abc.AsyncIterator: AsyncIterator,
    abc.Awaitable: Awaitable,
    **{union_t: Union for union_t in UnionTypes},
}
"""A mapping of types to equivalent types that are safe to be used as generics across all Python versions.

This is necessary because occasionally we want to rebuild a generic outer type with different args, and types such as
``collections.abc.Mapping``, are not valid generic types in Python 3.8.
"""

wrapper_type_set = {Annotated, Required, NotRequired}
"""Types that always contain a wrapped type annotation as their first arg."""


def normalize_type_annotation(annotation: Any) -> Any:
    """Normalize a type annotation to a standard form."""
    return instantiable_type_mapping.get(annotation, annotation)


def make_non_optional_union(annotation: UnionT | None) -> UnionT:
    """Make a :data:`Union <typing.Union>` type that excludes ``NoneType``.

    Args:
        annotation: A type annotation.

    Returns:
        The union with all original members, except ``NoneType``.
    """
    args = tuple(tp for tp in get_args(annotation) if tp is not NoneType)
    return cast("UnionT", Union[args])  # pyright: ignore


def unwrap_annotation(annotation: Any) -> tuple[Any, tuple[Any, ...], set[Any]]:
    """Remove "wrapper" annotation types, such as ``Annotated``, ``Required``, and ``NotRequired``.

    Note:
        ``annotation`` should have been retrieved from :func:`get_type_hints()` with ``include_extras=True``. This
        ensures that any nested ``Annotated`` types are flattened according to the PEP 593 specification.

    Args:
        annotation: A type annotation.

    Returns:
        A tuple of the unwrapped annotation and any ``Annotated`` metadata, and a set of any wrapper types encountered.
    """
    origin = get_origin(annotation)
    wrappers = set()
    metadata = []
    while origin in wrapper_type_set:
        wrappers.add(origin)
        annotation, *meta = get_args(annotation)
        metadata.extend(meta)
        origin = get_origin(annotation)
    return annotation, tuple(metadata), wrappers


def get_origin_or_inner_type(annotation: Any) -> Any:
    """Get origin or unwrap it. Returns None for non-generic types.

    Args:
        annotation: A type annotation.

    Returns:
        Any type.
    """
    origin = get_origin(annotation)
    if origin in wrapper_type_set:
        inner, _, _ = unwrap_annotation(annotation)
        # we need to recursively call here 'get_origin_or_inner_type' because we might be dealing
        # with a generic type alias e.g. Annotated[dict[str, list[int]]
        origin = get_origin_or_inner_type(inner)
    return instantiable_type_mapping.get(origin, origin)


def get_safe_generic_origin(origin_type: Any, annotation: Any) -> Any:
    """Get a type that is safe to use as a generic type across all supported Python versions.

    If a builtin collection type is annotated without generic args, e.g, ``a: dict``, then the origin type will be
    ``None``. In this case, we can use the annotation to determine the correct generic type, if one exists.

    Args:
        origin_type: A type - would be the return value of :func:`get_origin()`.
        annotation: Type annotation associated with the origin type. Should be unwrapped from any wrapper types, such
            as ``Annotated``.

    Returns:
        The ``typing`` module equivalent of the given type, if it exists. Otherwise, the original type is returned.
    """
    if origin_type is None:
        return safe_generic_origin_map.get(annotation)
    return safe_generic_origin_map.get(origin_type, origin_type)


def get_instantiable_origin(origin_type: Any, annotation: Any) -> Any:
    """Get a type that is safe to instantiate for the given origin type.

    If a builtin collection type is annotated without generic args, e.g, ``a: dict``, then the origin type will be
    ``None``. In this case, we can use the annotation to determine the correct instantiable type, if one exists.

    Args:
        origin_type: A type - would be the return value of :func:`get_origin()`.
        annotation: Type annotation associated with the origin type. Should be unwrapped from any wrapper types, such
            as ``Annotated``.

    Returns:
        A builtin type that is safe to instantiate for the given origin type.
    """
    if origin_type is None:
        return instantiable_type_mapping.get(annotation)
    return instantiable_type_mapping.get(origin_type, origin_type)


def get_type_hints_with_generics_resolved(
    annotation: Any,
    globalns: dict[str, Any] | None = None,
    localns: dict[str, Any] | None = None,
    include_extras: bool = False,
) -> dict[str, Any]:
    """Get the type hints for the given object after resolving the generic types as much as possible.

    Args:
        annotation: A type annotation.
        globalns: The global namespace.
        localns: The local namespace.
        include_extras: A flag indicating whether to include the ``Annotated[T, ...]`` or not.
    """
    origin = get_origin(annotation)

    if origin is None:
        # Implies the generic types have not been specified in the annotation
        type_hints = get_type_hints(annotation, globalns=globalns, localns=localns, include_extras=include_extras)
        typevar_map = {p: p for p in annotation.__parameters__}
    else:
        type_hints = get_type_hints(origin, globalns=globalns, localns=localns, include_extras=include_extras)
        # the __parameters__ is only available on the origin itself and not the annotation
        typevar_map = dict(zip(origin.__parameters__, get_args(annotation)))

    return {n: _substitute_typevars(type_, typevar_map) for n, type_ in type_hints.items()}


def _substitute_typevars(obj: Any, typevar_map: Mapping[Any, Any]) -> Any:
    if params := getattr(obj, "__parameters__", None):
        args = tuple(_substitute_typevars(typevar_map.get(p, p), typevar_map) for p in params)
        return obj[args]

    if isinstance(obj, TypeVar):
        # If there's a mapped type for the TypeVar already, then it should be returned instead
        # of considering __constraints__ or __bound__. For a generic `Foo[T]`, if Foo[int] is given
        # then int should be returned and if `Foo` is given then the __bounds__ and __constraints__
        # should be considered.
        if (type_ := typevar_map.get(obj, None)) is not None and not isinstance(type_, TypeVar):
            return type_

        if obj.__bound__ is not None:
            return obj.__bound__

        if obj.__constraints__:
            return Union[obj.__constraints__]  # pyright: ignore

    return obj



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from litestar import Litestar, MediaType, get
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.contrib.mako import MakoTemplateEngine
from litestar.contrib.minijinja import MiniJinjaTemplateEngine
from litestar.exceptions import ImproperlyConfiguredException
from litestar.response.template import Template
from litestar.template import TemplateEngineProtocol
from litestar.template.config import TemplateConfig
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from litestar import Request


def test_handler_raise_for_no_template_engine() -> None:
    @get(path="/")
    def invalid_path() -> Template:
        return Template(template_name="index.html", context={"ye": "yeeee"})

    with create_test_client(route_handlers=[invalid_path], debug=False) as client:
        response = client.request("GET", "/")
        assert response.status_code == 500
        assert response.json() == {"detail": "Internal Server Error", "status_code": 500}


def test_engine_passed_to_callback(tmp_path: Path) -> None:
    received_engine: JinjaTemplateEngine | None = None

    def callback(engine: TemplateEngineProtocol) -> None:
        nonlocal received_engine
        assert isinstance(engine, JinjaTemplateEngine), "Engine must be a JinjaTemplateEngine"
        received_engine = engine

    app = Litestar(
        route_handlers=[],
        template_config=TemplateConfig(
            directory=tmp_path,
            engine=JinjaTemplateEngine,
            engine_callback=callback,
        ),
    )

    assert received_engine is not None
    assert received_engine is app.template_engine


@pytest.mark.parametrize("engine", (JinjaTemplateEngine, MakoTemplateEngine, MiniJinjaTemplateEngine))
def test_engine_instance(engine: type[TemplateEngineProtocol], tmp_path: Path) -> None:
    engine_instance = engine(directory=tmp_path, engine_instance=None)
    if isinstance(engine_instance, JinjaTemplateEngine):
        assert engine_instance.engine.autoescape is True

    if isinstance(engine_instance, MakoTemplateEngine):
        assert engine_instance.engine.template_args["default_filters"] == ["h"]

    config = TemplateConfig(engine=engine_instance)
    assert config.engine_instance is engine_instance


@pytest.mark.parametrize("engine", (JinjaTemplateEngine, MakoTemplateEngine, MiniJinjaTemplateEngine))
def test_directory_validation(engine: type[TemplateEngineProtocol], tmp_path: Path) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        TemplateConfig(engine=engine)


@pytest.mark.parametrize("engine", (JinjaTemplateEngine, MakoTemplateEngine, MiniJinjaTemplateEngine))
def test_instance_and_directory_validation(engine: type[TemplateEngineProtocol], tmp_path: Path) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        TemplateConfig(engine=engine, instance=engine(directory=tmp_path, engine_instance=None))


@pytest.mark.parametrize("media_type", [MediaType.HTML, MediaType.TEXT, "text/arbitrary"])
def test_media_type(media_type: MediaType | str, tmp_path: Path) -> None:
    (tmp_path / "hello.tpl").write_text("hello")

    @get("/", media_type=media_type)
    def index() -> Template:
        return Template(template_name="hello.tpl")

    with create_test_client(
        [index], template_config=TemplateConfig(directory=tmp_path, engine=JinjaTemplateEngine)
    ) as client:
        res = client.get("/")
        assert res.status_code == 200
        assert res.headers["content-type"].startswith(
            media_type if isinstance(media_type, str) else media_type.value,  # type: ignore[union-attr]
        )


@pytest.mark.parametrize(
    "extension,expected_type",
    [
        (".json", MediaType.JSON),
        (".html", MediaType.HTML),
        (".html.other", MediaType.HTML),
        (".css", MediaType.CSS),
        (".xml", MediaType.XML),
        (".xml.other", MediaType.XML),
        (".txt", MediaType.TEXT),
        (".unknown", MediaType.TEXT),
        ("", MediaType.TEXT),
    ],
)
@pytest.mark.skipif(sys.platform == "win32", reason="mimetypes.guess_types is unreliable on windows")
def test_media_type_inferred(extension: str, expected_type: MediaType, tmp_path: Path) -> None:
    tpl_name = f"hello{extension}"
    (tmp_path / tpl_name).write_text("hello")

    @get("/")
    def index() -> Template:
        return Template(template_name=tpl_name)

    with create_test_client(
        [index], template_config=TemplateConfig(directory=tmp_path, engine=JinjaTemplateEngine)
    ) as client:
        res = client.get("/")
        assert res.status_code == 200
        assert res.headers["content-type"].startswith(expected_type.value)


def test_before_request_handler_content_type(tmp_path: Path) -> None:
    template_loc = tmp_path / "about.html"

    def before_request_handler(_: Request) -> None:
        template_loc.write_text("before request")

    @get("/", before_request=before_request_handler)
    def index() -> Template:
        return Template(template_name="about.html")

    with create_test_client(
        [index], template_config=TemplateConfig(directory=tmp_path, engine=JinjaTemplateEngine)
    ) as client:
        res = client.get("/")
        assert res.status_code == 200
        assert res.headers["content-type"].startswith(MediaType.HTML.value)
        assert res.text == "before request"


test_cases = [
    {"name": "both", "template_name": "dummy.html", "template_str": "Dummy", "raises": ValueError},
    {"name": "none", "template_name": None, "template_str": None, "status_code": 500},
    {"name": "name_only", "template_name": "dummy.html", "template_str": None, "status_code": 200},
    {"name": "str_only", "template_name": None, "template_str": "Dummy", "status_code": 200},
]


@pytest.mark.parametrize("engine", (JinjaTemplateEngine, MakoTemplateEngine, MiniJinjaTemplateEngine))
@pytest.mark.parametrize("test_case", test_cases, ids=[case["name"] for case in test_cases])  # type: ignore[index]
def test_template_scenarios(tmp_path: Path, engine: TemplateEngineProtocol, test_case: dict) -> None:
    if test_case["template_name"]:
        template_loc = tmp_path / test_case["template_name"]
        template_loc.write_text("Test content for template")

    @get("/")
    def index() -> Template:
        return Template(template_name=test_case["template_name"], template_str=test_case["template_str"])

    with create_test_client([index], template_config=TemplateConfig(directory=tmp_path, engine=engine)) as client:
        if "raises" in test_case and test_case["raises"] is ValueError:
            response = client.get("/")
            assert response.status_code == 500
            assert "ValueError" in response.text

        else:
            response = client.get("/")
            assert response.status_code == test_case["status_code"]

            if test_case["status_code"] == 200:
                if test_case["template_str"]:
                    assert response.text == test_case["template_str"]
                else:
                    assert response.text == "Test content for template"

            elif test_case["status_code"] == 500:
                assert "Either template_name or template_str must be provided" in response.text



from docs.examples.parameters.query_params import app as query_params_app
from docs.examples.parameters.query_params_constraints import (
    app as query_params_constraints_app,
)
from docs.examples.parameters.query_params_default import app as query_params_default_app
from docs.examples.parameters.query_params_optional import app as query_params_optional_app
from docs.examples.parameters.query_params_remap import app as query_params_remap_app
from docs.examples.parameters.query_params_types import app as query_params_types_app

from litestar.testing import TestClient


def test_query_params() -> None:
    with TestClient(app=query_params_app) as client:
        res = client.get("/?param=hello")
        assert res.status_code == 200
        assert res.json() == {"param": "hello"}


def test_query_params_default() -> None:
    with TestClient(app=query_params_default_app) as client:
        res = client.get("/")
        assert res.status_code == 200
        assert res.json() == {"param": "hello"}

        res = client.get("/?param=world")
        assert res.status_code == 200
        assert res.json() == {"param": "world"}


def test_query_params_optional() -> None:
    with TestClient(app=query_params_optional_app) as client:
        res = client.get("/")
        assert res.status_code == 200
        assert res.json() == {"param": None}

        res = client.get("/?param=world")
        assert res.status_code == 200
        assert res.json() == {"param": "world"}


def test_query_params_types() -> None:
    with TestClient(app=query_params_types_app) as client:
        res = client.get("/?date=2022-11-28T13:22:06.916540&floating_number=0.1&number=42&strings=1&strings=2")
        assert res.status_code == 200, res.json()
        assert res.json() == {"datetime": "2022-11-29T13:22:06.916540", "int": 42, "float": 0.1, "list": ["1", "2"]}


def test_query_params_remap() -> None:
    with TestClient(app=query_params_remap_app) as client:
        res = client.get("/?camelCase=hello")
        assert res.status_code == 200
        assert res.json() == {"param": "hello"}


def test_query_params_constraint() -> None:
    with TestClient(app=query_params_constraints_app) as client:
        res = client.get("/?param=1")
        assert res.status_code == 400

        res = client.get("/?param=6")
        assert res.status_code == 200
        assert res.json() == {"param": 6}



from docs.examples.dependency_injection import dependency_validation_error

from litestar.status_codes import HTTP_500_INTERNAL_SERVER_ERROR
from litestar.testing import TestClient


def test_route_returns_internal_server_error() -> None:
    with TestClient(app=dependency_validation_error.app) as client:
        r = client.get("/")
    assert r.status_code == HTTP_500_INTERNAL_SERVER_ERROR



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Tuple, Union

import pytest

from litestar.connection import ASGIConnection
from litestar.dto import DataclassDTO, DTOConfig, DTOField
from litestar.dto._backend import DTOBackend
from litestar.dto._types import (
    CollectionType,
    CompositeType,
    MappingType,
    SimpleType,
    TransferDTOFieldDefinition,
    TupleType,
    UnionType,
)
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.types import DataclassProtocol
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from typing import AbstractSet

    from litestar.dto._types import TransferType


@dataclass
class Model:
    a: int
    b: str


@dataclass
class Model2:
    c: int
    d: str


@pytest.fixture(name="data_model_type")
def fx_data_model_type() -> Any:
    return type("Model", (Model,), {})


@pytest.fixture(name="data_model")
def fx_data_model(data_model_type: type[Model]) -> Model:
    return data_model_type(a=1, b="2")


@pytest.fixture(name="field_definitions")
def fx_field_definitions(data_model_type: type[Model]) -> list[DTOFieldDefinition]:
    return [
        DTOFieldDefinition.from_field_definition(
            field_definition=FieldDefinition.from_kwarg(
                annotation=int,
                name="a",
            ),
            default_factory=None,
            dto_field=DTOField(),
            model_name="some_module.SomeModel",
        ),
        DTOFieldDefinition.from_field_definition(
            field_definition=FieldDefinition.from_kwarg(
                annotation=str,
                name="b",
            ),
            default_factory=None,
            dto_field=DTOField(),
            model_name="some_module.SomeModel",
        ),
    ]


@pytest.fixture(name="backend")
def fx_backend(
    data_model_type: type[Model],
    field_definitions: list[DTOFieldDefinition],
    backend_cls: type[DTOBackend],
    use_experimental_dto_backend: bool,
) -> DTOBackend:
    class _Factory(DataclassDTO):
        config = DTOConfig(experimental_codegen_backend=use_experimental_dto_backend)

        @classmethod
        def generate_field_definitions(
            cls, model_type: type[DataclassProtocol]
        ) -> Generator[DTOFieldDefinition, None, None]:
            yield from field_definitions

    class _Backend(backend_cls):  # type: ignore[valid-type,misc]
        def create_transfer_model_type(
            self, model_name: str, field_definitions: tuple[TransferDTOFieldDefinition, ...]
        ) -> type[Any]:
            """Create a model for data transfer.

            Args:
            unique_name: name for the type that should be unique across all transfer types.
            field_definitions: field definitions for the container type.

            Returns:
            A ``BackendT`` class.
            """
            return Model

        def parse_raw(self, raw: bytes, asgi_connection: ASGIConnection) -> Any:
            """Parse raw bytes into transfer model type.

            Args:
            raw: bytes
            asgi_connection: Information about the active connection.

            Returns:
            The raw bytes parsed into transfer model type.
            """
            return None

        def parse_builtins(self, builtins: Any, asgi_connection: ASGIConnection) -> Any:
            """Parse builtin types into transfer model type.

            Args:
            builtins: Builtin type.
            asgi_connection: Information about the active connection.

            Returns:
            The builtin type parsed into transfer model type.
            """
            return None

    return _Backend(
        dto_factory=_Factory,
        is_data_field=True,
        field_definition=FieldDefinition.from_annotation(data_model_type),
        model_type=data_model_type,
        wrapper_attribute_name=None,
        handler_id="test",
    )


def create_transfer_type(
    backend: DTOBackend,
    field_definition: FieldDefinition,
    exclude: AbstractSet[str] | None = None,
    include: AbstractSet[str] | None = None,
    field_name: str = "name",
    unique_name: str = "some_module.SomeModel.name",
    nested_depth: int = 0,
    rename_fields: dict[str, str] | None = None,
) -> TransferType:
    return backend._create_transfer_type(
        field_definition=field_definition,
        exclude=exclude or set(),
        include=include or set(),
        field_name=field_name,
        unique_name=unique_name,
        nested_depth=nested_depth,
        rename_fields=rename_fields or {},
    )


@pytest.mark.parametrize(
    ("field_definition", "should_have_nested", "has_nested_field_info"),
    [
        (FieldDefinition.from_annotation(Union[Model, None]), True, (True, False)),
        (FieldDefinition.from_annotation(Union[Model, str]), True, (True, False)),
        (FieldDefinition.from_annotation(Union[Model, int]), True, (True, False)),
        (FieldDefinition.from_annotation(Union[Model, Model2]), True, (True, True)),
        (FieldDefinition.from_annotation(Union[int, str]), False, (False, False)),
    ],
)
def test_create_transfer_type_union(
    field_definition: FieldDefinition,
    should_have_nested: bool,
    has_nested_field_info: tuple[bool, ...],
    backend: DTOBackend,
) -> None:
    transfer_type = create_transfer_type(backend, field_definition)
    assert isinstance(transfer_type, UnionType)
    assert transfer_type.has_nested is should_have_nested
    inner_types = transfer_type.inner_types
    assert len(inner_types) == len(transfer_type.field_definition.inner_types)
    for inner_type, has_nested in zip(inner_types, has_nested_field_info):
        assert isinstance(inner_type, SimpleType)
        assert bool(inner_type.nested_field_info) is has_nested


@pytest.mark.parametrize(
    ("field_definition", "should_have_nested", "has_nested_field_info"),
    [
        (FieldDefinition.from_annotation(Tuple[Model, None]), True, (True, False)),
        (FieldDefinition.from_annotation(Tuple[Model, str]), True, (True, False)),
        (FieldDefinition.from_annotation(Tuple[Model, int]), True, (True, False)),
        (FieldDefinition.from_annotation(Tuple[Model, Model2]), True, (True, True)),
        (FieldDefinition.from_annotation(Tuple[int, str]), False, (False, False)),
        (FieldDefinition.from_annotation(Tuple[Model, ...]), True, (True,)),
        (FieldDefinition.from_annotation(Tuple[int, ...]), False, (False,)),
    ],
)
def test_create_transfer_type_tuple(
    field_definition: FieldDefinition,
    should_have_nested: bool,
    has_nested_field_info: tuple[bool, ...],
    backend: DTOBackend,
) -> None:
    transfer_type = create_transfer_type(backend, field_definition)
    assert isinstance(transfer_type, CompositeType)
    assert transfer_type.has_nested is should_have_nested
    if field_definition.inner_types[-1].annotation is Ellipsis:
        assert isinstance(transfer_type, CollectionType)
        inner_type = transfer_type.inner_type
        assert isinstance(inner_type, SimpleType)
        assert bool(inner_type.nested_field_info) is has_nested_field_info[0]
    else:
        assert isinstance(transfer_type, TupleType)
        inner_types = transfer_type.inner_types
        assert len(inner_types) == len(transfer_type.field_definition.inner_types)
        for inner_type, has_nested in zip(inner_types, has_nested_field_info):
            assert isinstance(inner_type, SimpleType)
            assert bool(inner_type.nested_field_info) is has_nested


@pytest.mark.parametrize(
    ("field_definition", "should_have_nested", "has_nested_field_info"),
    [
        (FieldDefinition.from_annotation(Dict[Model, None]), True, (True, False)),
        (FieldDefinition.from_annotation(Dict[Model, str]), True, (True, False)),
        (FieldDefinition.from_annotation(Dict[Model, int]), True, (True, False)),
        (FieldDefinition.from_annotation(Dict[Model, Model2]), True, (True, True)),
        (FieldDefinition.from_annotation(Dict[int, str]), False, (False, False)),
    ],
)
def test_create_transfer_type_mapping(
    field_definition: FieldDefinition,
    should_have_nested: bool,
    has_nested_field_info: tuple[bool, ...],
    backend: DTOBackend,
) -> None:
    transfer_type = create_transfer_type(backend, field_definition)
    assert isinstance(transfer_type, MappingType)
    assert transfer_type.has_nested is should_have_nested
    key_type = transfer_type.key_type
    value_type = transfer_type.value_type
    for inner_type, has_nested in zip((key_type, value_type), has_nested_field_info):
        assert isinstance(inner_type, SimpleType)
        assert bool(inner_type.nested_field_info) is has_nested


@pytest.mark.parametrize(
    ("field_definition", "should_have_nested", "has_nested_field_info"),
    [
        (FieldDefinition.from_annotation(List[Model]), True, True),
        (FieldDefinition.from_annotation(List[int]), False, False),
    ],
)
def test_create_transfer_type_collection(
    field_definition: FieldDefinition,
    should_have_nested: bool,
    has_nested_field_info: bool,
    backend: DTOBackend,
) -> None:
    transfer_type = create_transfer_type(backend, field_definition)
    assert isinstance(transfer_type, CollectionType)
    assert transfer_type.has_nested is should_have_nested
    inner_type = transfer_type.inner_type
    assert isinstance(inner_type, SimpleType)
    assert bool(inner_type.nested_field_info) is has_nested_field_info


def test_create_collection_type_nested_union(backend: DTOBackend) -> None:
    field_definition = FieldDefinition.from_annotation(List[Union[Model, Model2]])
    transfer_type = create_transfer_type(backend, field_definition)
    assert isinstance(transfer_type, CollectionType)
    assert transfer_type.has_nested is True
    inner_type = transfer_type.inner_type
    assert isinstance(inner_type, UnionType)
    assert inner_type.has_nested is True
    inner_types = inner_type.inner_types
    assert len(inner_types) == len(inner_type.field_definition.inner_types)
    for inner_type in inner_types:
        assert isinstance(inner_type, SimpleType)
        assert bool(inner_type.nested_field_info)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, AsyncGenerator, AsyncIterable, AsyncIterator, Iterable, Iterator

from anyio.to_thread import run_sync

from litestar.exceptions import ImproperlyConfiguredException
from litestar.response.streaming import Stream
from litestar.utils import AsyncIteratorWrapper

if TYPE_CHECKING:
    from litestar.background_tasks import BackgroundTask, BackgroundTasks
    from litestar.types import ResponseCookies, ResponseHeaders, StreamType

_LINE_BREAK_RE = re.compile(r"\r\n|\r|\n")


class _ServerSentEventIterator(AsyncIteratorWrapper[bytes]):
    __slots__ = ("content_async_iterator",)

    content_async_iterator: AsyncIteratorWrapper[bytes | str] | AsyncIterable[str | bytes] | AsyncIterator[str | bytes]

    def __init__(
        self,
        content: str | bytes | StreamType[str | bytes],
        event_type: str | None = None,
        event_id: int | None = None,
        retry_duration: int | None = None,
        comment_message: str | None = None,
    ) -> None:
        chunks: list[bytes] = []
        if comment_message is not None:
            chunks.extend([f": {chunk}\r\n".encode() for chunk in _LINE_BREAK_RE.split(comment_message)])

        if event_id is not None:
            chunks.append(f"id: {event_id}\r\n".encode())

        if event_type is not None:
            chunks.append(f"event: {event_type}\r\n".encode())

        if retry_duration is not None:
            chunks.append(f"retry: {retry_duration}\r\n".encode())

        super().__init__(iterator=chunks)

        if not isinstance(content, (Iterator, AsyncIterator, AsyncIteratorWrapper)) and callable(content):
            content = content()  # type: ignore[unreachable]

        if isinstance(content, (str, bytes)):
            self.content_async_iterator = AsyncIteratorWrapper([content])
        elif isinstance(content, (Iterable, Iterator)):
            self.content_async_iterator = AsyncIteratorWrapper(content)  # type: ignore[arg-type]
        elif isinstance(content, (AsyncIterable, AsyncIterator, AsyncIteratorWrapper)):
            self.content_async_iterator = content
        else:
            raise ImproperlyConfiguredException(f"Invalid type {type(content)} for ServerSentEvent")

    def _call_next(self) -> bytes:
        try:
            return next(self.iterator)
        except StopIteration as e:
            raise ValueError from e

    async def _async_generator(self) -> AsyncGenerator[bytes, None]:
        while True:
            try:
                yield await run_sync(self._call_next)
            except ValueError:
                async for value in self.content_async_iterator:
                    yield f"data: {value}\r\n".encode() if isinstance(value, str) else b"data: {" + value + b"}\r\n"
                break
        yield b"\r\n"


class ServerSentEvent(Stream):
    def __init__(
        self,
        content: str | bytes | StreamType[str | bytes],
        *,
        background: BackgroundTask | BackgroundTasks | None = None,
        cookies: ResponseCookies | None = None,
        encoding: str = "utf-8",
        headers: ResponseHeaders | None = None,
        event_type: str | None = None,
        event_id: int | None = None,
        retry_duration: int | None = None,
        comment_message: str | None = None,
        status_code: int | None = None,
    ) -> None:
        """Initialize the response.

        Args:
            content: Bytes, string or a sync or async iterator or iterable.
            background: A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or
                :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
                Defaults to None.
            cookies: A list of :class:`Cookie <.datastructures.Cookie>` instances to be set under the response
                ``Set-Cookie`` header.
            encoding: The encoding to be used for the response headers.
            headers: A string keyed dictionary of response headers. Header keys are insensitive.
            status_code: The response status code. Defaults to 200.
            event_type: The type of the SSE event. If given, the browser will sent the event to any 'event-listener'
                declared for it (e.g. via 'addEventListener' in JS).
            event_id: The event ID. This sets the event source's 'last event id'.
            retry_duration: Retry duration in milliseconds.
            comment_message: A comment message. This value is ignored by clients and is used mostly for pinging.
        """
        super().__init__(
            content=_ServerSentEventIterator(
                content=content,
                event_type=event_type,
                event_id=event_id,
                retry_duration=retry_duration,
                comment_message=comment_message,
            ),
            media_type="text/event-stream",
            background=background,
            cookies=cookies,
            encoding=encoding,
            headers=headers,
            status_code=status_code,
        )
        self.headers.setdefault("Cache-Control", "no-cache")
        self.headers["Connection"] = "keep-alive"
        self.headers["X-Accel-Buffering"] = "no"



import gzip
import random
from datetime import timedelta
from typing import TYPE_CHECKING, Optional, Type, Union
from unittest.mock import MagicMock
from uuid import uuid4

import msgspec
import pytest

from litestar import Litestar, Request, Response, get, post
from litestar.config.compression import CompressionConfig
from litestar.config.response_cache import CACHE_FOREVER, ResponseCacheConfig
from litestar.enums import CompressionEncoding
from litestar.middleware.response_cache import ResponseCacheMiddleware
from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
from litestar.stores.base import Store
from litestar.stores.memory import MemoryStore
from litestar.testing import TestClient, create_test_client
from litestar.types import HTTPScope

if TYPE_CHECKING:
    from time_machine import Coordinates


@pytest.fixture()
def mock() -> MagicMock:
    return MagicMock(return_value=str(random.random()))


def after_request_handler(response: "Response") -> "Response":
    response.headers["unique-identifier"] = str(uuid4())
    return response


@pytest.mark.parametrize("sync_to_thread", (True, False))
def test_default_cache_response(sync_to_thread: bool, mock: MagicMock) -> None:
    @get(
        "/cached",
        sync_to_thread=sync_to_thread,
        cache=True,
        type_encoders={int: str},  # test pickling issues. see https://github.com/litestar-org/litestar/issues/1096
    )
    def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    with create_test_client([handler], after_request=after_request_handler) as client:
        first_response = client.get("/cached")
        second_response = client.get("/cached")

        first_response_identifier = first_response.headers["unique-identifier"]

        assert first_response.status_code == 200
        assert second_response.status_code == 200
        assert second_response.headers["unique-identifier"] == first_response_identifier
        assert first_response.text == second_response.text
        assert mock.call_count == 1


def test_handler_expiration(mock: MagicMock, frozen_datetime: "Coordinates") -> None:
    @get("/cached-local", cache=10)
    async def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    with create_test_client([handler], after_request=after_request_handler) as client:
        first_response = client.get("/cached-local")
        frozen_datetime.shift(delta=timedelta(seconds=5))
        second_response = client.get("/cached-local")
        assert first_response.headers["unique-identifier"] == second_response.headers["unique-identifier"]
        assert mock.call_count == 1

        frozen_datetime.shift(delta=timedelta(seconds=11))
        third_response = client.get("/cached-local")
        assert first_response.headers["unique-identifier"] != third_response.headers["unique-identifier"]
        assert mock.call_count == 2


def test_default_expiration(mock: MagicMock, frozen_datetime: "Coordinates") -> None:
    @get("/cached-default", cache=True)
    async def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    with create_test_client(
        [handler], after_request=after_request_handler, response_cache_config=ResponseCacheConfig(default_expiration=1)
    ) as client:
        first_response = client.get("/cached-default")
        second_response = client.get("/cached-default")
        assert first_response.headers["unique-identifier"] == second_response.headers["unique-identifier"]
        assert mock.call_count == 1

        frozen_datetime.shift(delta=timedelta(seconds=1))
        third_response = client.get("/cached-default")
        assert first_response.headers["unique-identifier"] != third_response.headers["unique-identifier"]
        assert mock.call_count == 2


@pytest.mark.parametrize("expiration,expected_expiration", [(True, None), (10, 10)])
def test_default_expiration_none(
    memory_store: MemoryStore, expiration: int, expected_expiration: Optional[int]
) -> None:
    @get("/cached", cache=expiration)
    def handler() -> None:
        return None

    app = Litestar(
        [handler],
        stores={"response_cache": memory_store},
        response_cache_config=ResponseCacheConfig(default_expiration=None),
    )

    with TestClient(app) as client:
        client.get("/cached")

    if expected_expiration is None:
        assert memory_store._store["GET/cached"].expires_at is None
    else:
        assert memory_store._store["GET/cached"].expires_at


def test_cache_forever(memory_store: MemoryStore) -> None:
    @get("/cached", cache=CACHE_FOREVER)
    async def handler() -> None:
        return None

    app = Litestar([handler], stores={"response_cache": memory_store})

    with TestClient(app) as client:
        client.get("/cached")

    assert memory_store._store["GET/cached"].expires_at is None


@pytest.mark.parametrize("sync_to_thread", (True, False))
async def test_custom_cache_key(sync_to_thread: bool, anyio_backend: str, mock: MagicMock) -> None:
    def custom_cache_key_builder(request: Request) -> str:
        return f"{request.url.path}:::cached"

    @get("/cached", sync_to_thread=sync_to_thread, cache=True, cache_key_builder=custom_cache_key_builder)
    def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    app = Litestar([handler])

    with TestClient(app) as client:
        client.get("/cached")
        store = app.stores.get("response_cache")
        assert await store.exists("/cached:::cached")


async def test_non_default_store_name(mock: MagicMock) -> None:
    @get(cache=True)
    def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    app = Litestar([handler], response_cache_config=ResponseCacheConfig(store="some_store"))

    with TestClient(app=app) as client:
        response_one = client.get("/")
        assert response_one.status_code == 200
        assert response_one.text == mock.return_value

        response_two = client.get("/")
        assert response_two.status_code == 200
        assert response_two.text == mock.return_value

        assert mock.call_count == 1

    assert await app.stores.get("some_store").exists("GET/")


async def test_with_stores(store: Store, mock: MagicMock) -> None:
    @get(cache=True)
    def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    app = Litestar([handler], stores={"response_cache": store})

    with TestClient(app=app) as client:
        response_one = client.get("/")
        assert response_one.status_code == 200
        assert response_one.text == mock.return_value

        response_two = client.get("/")
        assert response_two.status_code == 200
        assert response_two.text == mock.return_value

        assert mock.call_count == 1


def test_does_not_apply_to_non_cached_routes(mock: MagicMock) -> None:
    @get("/")
    def handler() -> str:
        return mock()  # type: ignore[no-any-return]

    with create_test_client([handler]) as client:
        first_response = client.get("/")
        second_response = client.get("/")

        assert first_response.status_code == 200
        assert second_response.status_code == 200
        assert mock.call_count == 2


@pytest.mark.parametrize(
    "cache,expect_applied",
    [
        (True, True),
        (False, False),
        (1, True),
        (CACHE_FOREVER, True),
    ],
)
def test_middleware_not_applied_to_non_cached_routes(
    cache: Union[bool, int, Type[CACHE_FOREVER]], expect_applied: bool
) -> None:
    @get(path="/", cache=cache)
    def handler() -> None:
        ...

    client = create_test_client(route_handlers=[handler])
    unpacked_middleware = []
    cur = client.app.asgi_router.root_route_map_node.children["/"].asgi_handlers["GET"][0]
    while hasattr(cur, "app"):
        unpacked_middleware.append(cur)
        cur = cur.app
    unpacked_middleware.append(cur)

    assert len([m for m in unpacked_middleware if isinstance(m, ResponseCacheMiddleware)]) == int(expect_applied)


async def test_compression_applies_before_cache() -> None:
    return_value = "_litestar_" * 4000
    mock = MagicMock(return_value=return_value)

    @get(path="/", cache=True)
    def handler_fn() -> str:
        return mock()  # type: ignore[no-any-return]

    app = Litestar(
        route_handlers=[handler_fn],
        compression_config=CompressionConfig(backend="gzip"),
    )

    with TestClient(app) as client:
        client.get("/", headers={"Accept-Encoding": str(CompressionEncoding.GZIP.value)})

    stored_value = await app.response_cache_config.get_store_from_app(app).get("GET/")
    assert stored_value
    stored_messages = msgspec.msgpack.decode(stored_value)
    assert gzip.decompress(stored_messages[1]["body"]).decode() == return_value


@pytest.mark.parametrize(
    ("response", "should_cache"),
    [
        (HTTP_200_OK, True),
        (HTTP_400_BAD_REQUEST, False),
        (HTTP_500_INTERNAL_SERVER_ERROR, False),
        (RuntimeError, False),
    ],
)
def test_default_do_response_cache_predicate(
    mock: MagicMock, response: Union[int, Type[RuntimeError]], should_cache: bool
) -> None:
    @get("/", cache=True)
    def handler() -> Response:
        mock()
        if isinstance(response, int):
            return Response(None, status_code=response)
        raise RuntimeError

    with create_test_client([handler]) as client:
        client.get("/")
        client.get("/")
        assert mock.call_count == 1 if should_cache else 2


def test_custom_do_response_cache_predicate(mock: MagicMock) -> None:
    @get("/", cache=True)
    def handler() -> str:
        mock()
        return "OK"

    def filter_cache_response(_: HTTPScope, __: int) -> bool:
        return False

    with create_test_client(
        [handler], response_cache_config=ResponseCacheConfig(cache_response_filter=filter_cache_response)
    ) as client:
        client.get("/")
        client.get("/")
        assert mock.call_count == 2


def test_on_multiple_handlers(mock: MagicMock) -> None:
    @get("/cached-local", cache=10)
    async def handler() -> str:
        mock()
        return "get_response"

    @post("/cached-local", cache=10)
    async def handler_post() -> str:
        mock()
        return "post_response"

    with create_test_client([handler, handler_post], after_request=after_request_handler) as client:
        # POST request to have this cached
        first_post_response = client.post("/cached-local")
        assert first_post_response.status_code == HTTP_201_CREATED
        assert first_post_response.text == "post_response"
        assert mock.call_count == 1

        # GET request to verify it doesn't use the cache created by the previous POST request
        get_response = client.get("/cached-local")
        assert get_response.status_code == HTTP_200_OK
        assert get_response.text == "get_response"
        assert first_post_response.headers["unique-identifier"] != get_response.headers["unique-identifier"]
        assert mock.call_count == 2

        # POST request to verify it uses the cache generated during the initial POST request
        second_post_response = client.post("/cached-local")
        assert second_post_response.status_code == HTTP_201_CREATED
        assert second_post_response.text == "post_response"
        assert first_post_response.headers["unique-identifier"] == second_post_response.headers["unique-identifier"]
        assert mock.call_count == 2



from typing import AsyncGenerator, Generator
from unittest.mock import MagicMock

import pytest

from litestar._kwargs.cleanup import DependencyCleanupGroup
from litestar.utils.compat import async_next


@pytest.fixture
def cleanup_mock() -> MagicMock:
    return MagicMock()


@pytest.fixture
def async_cleanup_mock() -> MagicMock:
    return MagicMock()


@pytest.fixture
def generator(cleanup_mock: MagicMock) -> Generator[str, None, None]:
    def func() -> Generator[str, None, None]:
        yield "hello"
        cleanup_mock()

    return func()


@pytest.fixture
def async_generator(async_cleanup_mock: MagicMock) -> AsyncGenerator[str, None]:
    async def func() -> AsyncGenerator[str, None]:
        yield "world"
        async_cleanup_mock()

    return func()


def test_add(generator: Generator[str, None, None]) -> None:
    group = DependencyCleanupGroup()

    group.add(generator)

    assert group._generators == [generator]


async def test_cleanup(generator: Generator[str, None, None], cleanup_mock: MagicMock) -> None:
    next(generator)
    group = DependencyCleanupGroup([generator])

    await group.cleanup()

    cleanup_mock.assert_called_once()
    assert group._closed


async def test_cleanup_multiple(
    generator: Generator[str, None, None],
    async_generator: AsyncGenerator[str, None],
    cleanup_mock: MagicMock,
    async_cleanup_mock: MagicMock,
) -> None:
    next(generator)
    await async_next(async_generator)
    group = DependencyCleanupGroup([generator, async_generator])

    await group.cleanup()

    cleanup_mock.assert_called_once()
    async_cleanup_mock.assert_called_once()
    assert group._closed


async def test_cleanup_on_closed_raises(generator: Generator[str, None, None]) -> None:
    next(generator)
    group = DependencyCleanupGroup([generator])

    await group.cleanup()
    with pytest.raises(RuntimeError):
        await group.cleanup()


async def test_add_on_closed_raises(
    generator: Generator[str, None, None], async_generator: AsyncGenerator[str, None]
) -> None:
    next(generator)
    group = DependencyCleanupGroup([generator])

    await group.cleanup()

    with pytest.raises(RuntimeError):
        group.add(async_generator)
